#! /usr/bin/env python
#
#   gsentitweet.py -- Front-end script for evaluating running multiple
#   Vader sentiment analysis jobs from an initial list of twitter archive
#   documents.
#
#   Copyright (C) 2017, 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running multiple Vader sentiment analysis jobs
from an initial list of twitter archive documents.

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsentitweet.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2017-06-26:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gsentitweet
    gsentitweet.gsentitweetScript().run()

import os
import sys
import time
import tarfile
import zipfile

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

COMMAND = "python twitter-senti-extract.py {twitter_file} {cores}"

## Utility methods
def _get_twitter_data(input_folder):
    """
    Returns list of valid .tar/.zip input files.
    """

    return [ os.path.join(input_folder,valid_input_files) for valid_input_files in os.listdir(input_folder) if tarfile.is_tarfile(os.path.join(input_folder,
                                                                                                                    valid_input_files)) or zipfile.is_zipfile(os.path.join(input_folder,
                                                                                                                                                                          valid_input_files)) ]

## custom application class
class gsentitweetApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gsentitweet'

    def __init__(self, input_file, getsentiment_script, cores, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        if not extra_args['sharedFS']:
            inputs[input_file] = os.path.basename(input_file)
            cmd = COMMAND.format(twitter_file=inputs[input_file], cores=cores)
        else:
            cmd = COMMAND.format(twitter_file=input_file, cores=cores)

        if getsentiment_script:
            inputs[getsentiment_script] = "./twitter-senti-extract.py"
        else:
            wrapper = resource_filename(Requirement.parse("gc3pie"),
                                        "gc3libs/etc/twitter-senti-extract.py")
            inputs[wrapper] = "./twitter-senti-extract.py"

        extra_args['requested_cores'] = cores

        Application.__init__(
            self,
            arguments = cmd,
            inputs = inputs,
            outputs = ["{input_file_name_prefix}.csv".format(input_file_name_prefix=os.path.basename(input_file))],
            stdout = 'gsentitweet.log',
            join=True,
            executables = executables,
            **extra_args)

class gsentitweetScript(SessionBasedScript):
    """
    The ``gsentitweet`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsentitweet`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gsentitweet``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = gsentitweetApplication,
            stats_only_for = gsentitweetApplication,
            )

    def setup_options(self):
        self.add_param("-P", "--getsentiment", metavar="[PATH]",
                       type=existing_file,
                       dest="getsentiment_script", default=None,
                       help="Location of python script to process input twitter file. "
                       "Default: %(default)s.")

        self.add_param("-S", "--sharedfs", dest="shared_FS",
                       action="store_true", default=True,
                       help="Whether the destination resource should assume shared filesystem where Input/Output data will be made available. Data transfer will happen through lcoal filesystem. Default: %(default)s.")

        self.add_param("--cores", dest="core_count", default=4, help="Specifiy number of cores to use Default: %(default)s.")

    def setup_args(self):
        self.add_param('input_folder',
                       type=existing_directory,
                       help="Path to input folder containing valid input "
                       " twitter .tar/zip files. Does NOT navigate the folder.")

    def new_tasks(self, extra):
        """
        For each valid input file create a new gsentitweetRetryableTask
        """
        tasks = []

        for input_file in _get_twitter_data(self.params.input_folder):
            jobname = os.path.basename(input_file)

            extra_args = extra.copy()
            extra_args['jobname'] = jobname
            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)
            extra_args['sharedFS'] = self.params.shared_FS

            self.log.debug("Creating Application for twitter data '%s'" % (input_file))

            tasks.append(gsentitweetApplication(
                input_file,
                self.params.getsentiment_script,
                self.params.core_count,
                **extra_args))

        return tasks



#!/usr/bin/env python
#
"""
Implementation of the command-line front-ends.
"""
# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
from __future__ import absolute_import, print_function, unicode_literals

from builtins import str
__docformat__ = 'reStructuredText'
__author__ = ', '.join([
    "Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>",
    "Riccardo Murri <riccardo.murri@uzh.ch>",
    "Antonio Messina <antonio.messina@uzh.ch>",
])

# stdlib imports
import csv
import sys
import os
import posix
import time
import types
import re
import multiprocessing as mp

# 3rd party modules
from parsedatetime.parsedatetime import Calendar
from prettytable import PrettyTable

# local modules
from gc3libs import __version__, Run
import gc3libs.defaults
from gc3libs.quantity import Duration, Memory
from gc3libs.session import Session, TemporarySession
import gc3libs.cmdline
import gc3libs.exceptions
import gc3libs.persistence
from gc3libs.url import Url
import gc3libs.utils as utils


class GC3UtilsScript(gc3libs.cmdline._Script):
    """
    Base class for GC3Utils scripts.

    The default command line implemented is the following:

      script [options] JOBID [JOBID ...]

    By default, only the standard options ``-h``/``--help`` and
    ``-V``/``--version`` are considered; to add more, override
    `setup_options`:meth:
    To change default positional argument parsing, override
    `setup_args`:meth:

    """

    def __init__(self, **extra_args):
        """
        Set version string in `gc3libs.cmdline._Script`:class: to
        this package's version.
        """
        super(GC3UtilsScript, self).__init__(
            main=self.main,
            version=__version__,
            name='gc3utils',  # for logging purposes
            **extra_args)

    ##
    # CUSTOMIZATION METHODS
    ##
    # The following are meant to be freely customized in derived scripts.
    ##

    def setup_args(self):
        """
        Set up command-line argument parsing.

        The default command line parsing considers every argument as a
        job ID; actual processing of the IDs is done in
        `parse_args`:meth:
        """
        self.add_param(
            'args',
            nargs='*',
            metavar='JOBID',
            help="Job ID string identifying the jobs to operate upon.")

    def parse_args(self):
        if hasattr(self.params, 'args') and '-' in self.params.args:
            # Get input arguments *also* from standard input
            self.params.args.remove('-')
            self.params.args.extend(sys.stdin.read().split())

    ##
    # pyCLI INTERFACE METHODS
    ##
    # The following methods adapt the behavior of the
    # `SessionBasedScript` class to the interface expected by pyCLI
    # applications.  Think twice before overriding them, and read
    # the pyCLI docs before :-)
    ##

    def setup(self):
        """
        Setup standard command-line parsing.

        GC3Utils scripts should probably override `setup_args`:meth:
        and `setup_options`:meth: to modify command-line parsing.
        """
        # setup of base classes (creates the argparse stuff)
        super(GC3UtilsScript, self).setup()
        # local additions
        self.add_param("-s",
                       "--session",
                       action="store",
                       required=True,
                       default=gc3libs.defaults.JOBS_DIR,
                       help="Directory where job information will be stored.")

    def pre_run(self):
        """
        Perform parsing of standard command-line options and call into
        `parse_args()` to do non-optional argument processing.
        """
        # base class parses command-line
        super(GC3UtilsScript, self).pre_run()

    ##
    # INTERNAL METHODS
    ##
    # The following methods are for internal use; they can be
    # overridden and customized in derived classes, although there
    # should be no need to do so.
    ##

    def _get_tasks(self, task_ids, ignore_failures=True):
        """
        Iterate over tasks (gc3libs.Application objects) corresponding
        to the given IDs.

        If `ignore_failures` is `True` (default), errors retrieving a
        job from the persistence layer are ignored and the jobid is
        skipped, therefore the returned list can be shorter than the
        list of Job IDs given as argument.  If `ignore_failures` is
        `False`, then any errors result in the relevant exception being
        re-raised.
        """
        for jobid in task_ids:
            try:
                yield self.session.load(jobid)
            except Exception as ex:
                # Exempted from GC3Pie's `error_ignored()` policy as there
                # is explicit control via the `ignore_failures` parameter
                if ignore_failures:
                    gc3libs.log.error(
                        "Could not retrieve job '%s' (%s: %s). Ignoring.",
                        jobid, ex.__class__.__name__, ex,
                        exc_info=(self.params.verbose > 2))
                    continue
                else:
                    raise

    def _get_session(self, url, **extra_args):
        """
        Return a `gc3libs.session.Session` object corresponding to the
        session identified by `url`.

        Any keyword arguments will be passed unchanged to the session
        object constructor.

        :raise gc3libs.exceptions.InvalidArgument:
          If the session cannot be loaded (e.g., does not exist).
        """
        try:
            url = Url(url)
            if url.scheme == 'file':
                extra_args.setdefault('create', False)
                return Session(url.path, **extra_args)
            else:
                return TemporarySession(url, **extra_args)
        except gc3libs.exceptions.InvalidArgument as err:
            raise RuntimeError(
                "Cannot load session `{0}`: {1}".format(url, err))

    def _list_all_tasks(self):
        try:
            return self.session.store.list()
        except NotImplementedError:
            raise NotImplementedError(
                "Task storage module does not allow listing all tasks."
                " Please specify the task IDs you wish to operate on.")


# ====== Main ========


class cmd_gclean(GC3UtilsScript):
    """
Permanently remove jobs from local and remote storage.

In normal operation, only jobs that are in a terminal status can
be removed; if you want to force ``gclean`` to remove a job that
is not in any one of those states, add the ``-f`` option to the
command line.

If a job description cannot be successfully read, the
corresponding job will not be deleted; use the ``-f`` option to
force removal of a job regardless.
    """

    def setup_options(self):
        self.add_param("-A",
                       action="store_true",
                       dest="all",
                       default=False,
                       help="Remove all stored jobs.")
        self.add_param("-f", "--force",
                       action="store_true",
                       dest="force",
                       default=False,
                       help="Remove job even when not in terminal state.")

    def main(self):
        if self.params.all and len(self.params.args) > 0:
            raise gc3libs.exceptions.InvalidUsage(
                "Option '-A' conflicts with list of job IDs to remove.")

        if self.params.all:
            self.session = self._get_session(self.params.session)
            self.params.args = self._list_all_tasks()
            if len(args) == 0:
                self.log.info("No jobs in session: nothing to do.")
        else:
            if len(self.params.args) == 0:
                self.log.warning(
                    "No job IDs given on command line: nothing to do."
                    " Type '%s --help' for usage help."
                    # if we were called with an absolute path,
                    # presume the command has been found by the
                    # shell through PATH and just print the command name,
                    # otherwise print the exact path name.
                    % (os.path.basename(sys.argv[0])
                       if os.path.isabs(sys.argv[0])
                       else sys.argv[0]))
                return os.EX_USAGE
            self.session = self._get_session(self.params.session,
                                             task_ids=self.params.args)

        failed = 0
        for jobid in self.params.args:
            try:
                app = self.session.store.load(jobid)
                app.attach(self._core)

                if app.execution.state != Run.State.NEW:
                    if app.execution.state not in [Run.State.TERMINATED,
                                                   Run.State.TERMINATING]:
                        if self.params.force:
                            self.log.warning(
                                "Job '%s' not in terminal state:"
                                " attempting to kill before cleaning up.", app)
                            try:
                                app.kill()
                            except Exception as ex:
                                self.log.warning(
                                    "Killing job '%s' failed (%s: %s);"
                                    " continuing anyway, but errors might"
                                    " ensue.",
                                    app, ex.__class__.__name__, str(ex))

                                app.execution.state = Run.State.TERMINATED
                        else:
                            failed += 1
                            self.log.error(
                                "Job '%s' not in terminal state: ignoring.",
                                app)
                            continue

                    try:
                        app.free()
                    except Exception as ex:
                        if self.params.force:
                            pass
                        else:
                            failed += 1
                            self.log.warning(
                                "Freeing job '%s' failed (%s: %s);"
                                " continuing anyway, but errors might ensue.",
                                app, ex.__class__.__name__, str(ex))
                            continue

            except gc3libs.exceptions.LoadError:
                if self.params.force:
                    pass
                else:
                    failed += 1
                    self.log.error("Could not load '%s': ignoring"
                                   " (use option '-f' to remove regardless).",
                                   jobid)
                    continue

            try:
                if jobid in self.session.tasks:
                    self.session.remove(jobid)
                else:
                    # if jobid is not a toplevel job Session.remove()
                    # will raise an error.
                    self.session.store.remove(jobid)
                self.log.info("Removed job '%s'", jobid)
            except:
                failed += 1
                self.log.error("Failed removing '%s' from persistency layer."
                               " option '-f' harmless" % jobid)
                continue

        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gclient(gc3libs.cmdline.DaemonClient):
    """
    Generic client for GC3Pie's XML-RPC server.

    Use the `help` command to display a list of remote functions
    supported by the server.
    """
    # all the functionality needed is already in the base class, no
    # need to add anything here
    pass


class cmd_ginfo(GC3UtilsScript):
    """
Print detailed information about a job.

A complete dump of all the information known about jobs listed on
the command line is printed; this will only make sense if you know
GC3Libs internals.
    """

    verbose_logging_threshold = 2

    def setup_options(self):
        self.add_param("-c", "--csv", action="store_true", dest="csv",
                       default=False,
                       help="Print attributes in CSV format,"
                       " e.g., for generating files that can be"
                       " read by a spreadsheet program."
                       " MUST be used together with '--print'.")
        self.add_param("--no-header",
                       action="store_false", dest="header", default=True,
                       help="Do *not* print table or CSV file header.")
        self.add_param("-p", "--print", action="store", dest="keys",
                       metavar="LIST", default='',
                       help="Only print job attributes whose name appears in"
                       " this comma-separated list.")
        self.add_param("-t", "--tabular", action="store_true", dest="tabular",
                       default=False,
                       help="Print attributes in table format."
                       " MUST be used together with '--print'.")
        self.add_param("-H", "--history", action="store_true", dest="history",
                       default=False,
                       help="Print job history only")

    def main(self):
        # sanity checks for the command-line options
        if self.params.csv and self.params.tabular:
            raise gc3libs.exceptions.InvalidUsage(
                "Conflicting options `-c`/`--csv` and `-t`/`--tabular`."
                " Choose either one, but not both.")

        if self.params.history and self.params.tabular:
            raise gc3libs.exceptions.InvalidUsage(
                "Conflicting options `-H`/`--history` and `-t`/`--tabular`."
                " Choose either one, but not both.")

        if self.params.history and self.params.csv:
            raise gc3libs.exceptions.InvalidUsage(
                "Conflicting options `-H`/`--history` and `-c`/`--csv`."
                " Choose either one, but not both.")

        if ((self.params.tabular or self.params.csv)
                and len(self.params.keys) == 0):
            raise gc3libs.exceptions.InvalidUsage(
                "Options '--tabular' and `--csv` only make sense"
                " in conjuction with option '--print'.")

        if posix.isatty(sys.stdout.fileno()):
            # try to screen width
            try:
                width = int(os.environ['COLUMNS'])
            except:
                width = 0
        else:
            # presume output goes to a file, so no width restrictions
            width = 0

        # start reading stuff from disk
        self.session = self._get_session(self.params.session,
                                         task_ids=self.params.args)

        if len(self.params.args) == 0:
            # if no arguments, operate on all known jobs
            self.params.args = self._list_all_tasks()

        if len(self.params.keys) > 0:
            only_keys = self.params.keys.split(',')
        else:
            if self.params.verbose < 2:
                def names_not_starting_with_underscore(name):
                    return not name.startswith('_')
                only_keys = names_not_starting_with_underscore
            else:
                # print *all* keys if `-vv` is given
                only_keys = None

        if self.params.tabular:
            # prepare table prettyprinter
            table = PrettyTable()
            table.border = True
            if self.params.header:
                table.field_names = ["Job ID"] + only_keys
                table.align = 'l'

        if self.params.csv:
            csv_output = csv.writer(sys.stdout)
            if self.params.header:
                csv_output.writerow(only_keys)

        ok = 0
        for app in sorted(self._get_tasks(self.params.args),
                          key=(lambda task: task.persistent_id)):
            # since `_get_tasks` swallows any exception raised by
            # invalid job IDs or corrupted files, let us determine the
            # number of failures by counting the number of times we
            # actually run this loop and then subtract from the number
            # of times we *should* have run, i.e., the number of
            # arguments we were passed.
            ok += 1
            if self.params.tabular or self.params.csv:
                row = [str(app)]
                for key in only_keys:
                    try:
                        row.append(gc3libs.utils.getattr_nested(app, key))
                    except AttributeError:
                        row.append("N/A")
                if self.params.tabular:
                    table.add_row(row)
                elif self.params.csv:
                    csv_output.writerow(row)
            elif self.params.history:
                print(str(app.persistent_id))
                indent = 4*' '
                for logentry in app.execution.history:
                    print(indent + logentry)
            else:
                # usual YAML-like output
                print(str(app.persistent_id))
                if self.params.verbose == 0:
                    utils.prettyprint(app.execution, indent=4,
                                      width=width, only_keys=only_keys)
                else:
                    # with `-v` and above, dump the whole `Application` object
                    utils.prettyprint(app, indent=4, width=width,
                                      only_keys=only_keys)
        if self.params.tabular:
            print(table)
        failed = len(self.params.args) - ok
        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gresub(GC3UtilsScript):
    """
Resubmit an already-submitted job with (possibly) different parameters.

If you resubmit a job that is not in terminal state, the existing job
is canceled before re-submission.
    """

    def setup_options(self):
        self.add_param("-r", "--resource",
                       action="store",
                       dest="resource_name",
                       metavar="NAME",
                       default=None,
                       help='Select execution resource by name')
        self.add_param("-c", "--cores",
                       action="store",
                       dest="ncores",
                       type=int,
                       metavar="NUM",
                       help='Request running job on this number of CPU cores')
        self.add_param("-m", "--memory",
                       action="store",
                       dest="memory_per_core",
                       metavar="GIGABYTES",
                       help="Set the amount of memory required per execution"
                       " core; default: %(default)s. Specify this as an"
                       " integral number followed by a unit, e.g., '512MB'"
                       " or '4GB'. execution site")
        self.add_param("-w", "--walltime",
                       action="store",
                       dest="walltime",
                       metavar="DURATION",
                       help="Set the time limit for each job."
                       " Jobs exceeding this limit will be stopped and"
                       " considered as 'failed'. The duration can be expressed"
                       " as a whole number followed by a time unit, e.g.,"
                       " '3600 s', '60 minutes', '8 hours', or a combination"
                       " thereof, e.g., '2hours 30minutes'.")

    def parse_args(self):
        if self.params.walltime:
            self.params.walltime = Duration(self.params.walltime)

        if self.params.memory_per_core:
            self.params.memory_per_core = Memory(self.params.memory_per_core)

    def main(self):
        if len(self.params.args) == 0:
            self.log.error(
                "No job IDs given on command line: nothing to do."
                " Type '%s --help' for usage help."
                # if we were called with an absolute path,
                # presume the command has been found by the
                # shell through PATH and just print the command name,
                # otherwise print the exact path name.
                % (os.path.basename(sys.argv[0]) if os.path.isabs(sys.argv[0])
                   else sys.argv[0]))
            return os.EX_USAGE

        if self.params.resource_name:
            self._select_resources(self.params.resource_name)
            self.log.info(
                "Retained only resources: %s (restricted by command-line"
                " option '-r %s')",
                str.join(",",
                         [res['name'] for res in self._core.get_resources()]),
                self.params.resource_name)

        self.session = self._get_session(self.params.session,
                                         task_ids=self.params.args)
        failed = 0
        for jobid in self.params.args:
            app = self.session.load(jobid.strip())

            # Update the requested walltime, memory per core and
            # number of cores for the application
            if self.params.ncores:
                app.requested_cores = self.params.ncores
            if self.params.memory_per_core:
                app.requested_memory = self.params.memory_per_core\
                    * app.requested_cores
            if self.params.walltime:
                app.requested_walltime = self.params.walltime
            app.attach(self._core)
            try:
                app.update_state()  # update state
            except Exception as ex:
                # ignore errors, and proceed to resubmission anyway
                self.log.warning("Could not update state of %s: %s: %s",
                                 jobid, ex.__class__.__name__, str(ex))
            # kill remote job
            try:
                app.kill()
            except Exception as ex:
                # ignore errors (alert user?)
                pass

            try:
                app.submit(resubmit=True)
                print("Successfully re-submitted %s; use the 'gstat' command"
                      " to monitor its progress." % app)
                self.session.store.replace(jobid, app)
            except Exception as ex:
                failed += 1
                self.log.error("Failed resubmission of job '%s': %s: %s",
                               jobid, ex.__class__.__name__, str(ex))

        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gstat(GC3UtilsScript):
    """
Print job state.
    """
    verbose_logging_threshold = 1

    def setup_options(self):
        self.add_param("-b", "--brief", "--summary",
                       action="store_true",
                       dest="summary",
                       help=("Only print a summary table"
                             " with count of jobs per each state."))
        self.add_param("-l", "--state",
                       action="store",
                       dest="states",
                       metavar="STATE",
                       default=None,
                       help="Only report about jobs in the given state."
                       " Multiple states are allowed: separate them with"
                       " commas.")
        self.add_param("-L", "--lifetimes", "--print-lifetimes",
                       nargs='?',
                       metavar='FILE',
                       action='store',
                       dest='lifetimes',
                       default=None,  # no option and no argument: discard data
                       const=sys.stdout,  # option given, but no argument
                       help="For each successful job, print"
                       " submission, start, and duration times."
                       " If FILE is omitted, report is printed to screen.")
        self.add_param("-u", "--update",
                       action="store_true",
                       dest="update",
                       default=False,
                       help="Update job statuses before printing results")
        self.add_param("-p", "--print",
                       action="store",
                       dest="keys",
                       metavar="LIST",
                       default=None,
                       help="Additionally print job attributes whose name"
                       " appears in this comma-separated list.")

    def main(self):
        # by default, DO NOT update job statuses

        if len(self.params.args) == 0:
            # if no arguments, operate on all known jobs
            self.session = self._get_session(self.params.session)
            self.params.args = self._list_all_tasks()
        else:
            self.session = self._get_session(self.params.session,
                                             task_ids=self.params.args)

        if len(self.params.args) == 0:
            print("No jobs submitted.")
            return 0

        if posix.isatty(sys.stdout.fileno()):
            # try to determine how many lines of output can we fit in a screen
            try:
                capacity = int(os.environ['LINES']) - 5
            except:
                capacity = 20
        else:
            # output is dumped to a file, so no restrictions
            capacity = gc3libs.utils.PlusInfinity

        # limit to specified job states?
        if self.params.states is not None:
            states = self.params.states.split(',')
        else:
            states = None

        # any additional values to print?
        if self.params.keys is not None:
            keys = self.params.keys.split(',')
        else:
            keys = []

        # init lifetimes report (if requested)
        if self.params.lifetimes is not None:
            if isinstance(self.params.lifetimes, (str,)):
                self.params.lifetimes = open(self.params.lifetimes, 'w')
            lifetimes_rows = [['JOBID',
                               'SUBMITTED_AT',
                               'RUNNING_AT',
                               'FINISHED_AT',
                               'WAIT_DURATION',
                               'EXEC_DURATION']]

        # update states and compute statistics
        stats = utils.defaultdict(lambda: 0)
        tot = 0
        rows = []
        for app in self._get_tasks(self.params.args):
            tot += 1  # one more job successfully loaded
            jobid = app.persistent_id
            if self.params.update:
                app.attach(self._core)
                app.update_state()
                self.session.store.replace(jobid, app)
            if states is None or app.execution.in_state(*states):
                try:
                    jobname = app.jobname
                except AttributeError:
                    jobname = ''

                key_values = []
                for name in keys:
                    if name in app.execution:
                        key_values.append(app.execution.get(name))
                    else:
                        key_values.append(app.get(name, "N/A"))
                rows.append([jobid,
                             jobname,
                             app.execution.state,
                             app.execution.info] + key_values)

            stats[app.execution.state] += 1
            if app.execution.state == Run.State.TERMINATED:
                if app.execution.returncode == 0:
                    stats['ok'] += 1
                    if self.params.lifetimes is not None:
                        try:
                            timestamps = app.execution.timestamp
                        except AttributeError:
                            # missing .execution or .terminated: job
                            # is malformed, skip it
                            continue
                        if Run.State.SUBMITTED in timestamps:
                            submitted_at = timestamps['SUBMITTED']
                        else:
                            # skip job: not enough info
                            continue
                        if Run.State.RUNNING in timestamps:
                            running_at = timestamps[Run.State.RUNNING]
                        else:
                            # this means that the transition from
                            # SUBMITTED to RUNNING to TERMINATED
                            # happened in between two updates; we have
                            # no idea what the update cycle was, so...
                            # there's nothing left to do but skip this job
                            continue
                        terminated_at = timestamps['TERMINATING']
                        lifetimes_rows.append(
                            [jobid,
                             submitted_at,
                             running_at,
                             terminated_at,
                             running_at - submitted_at,
                             terminated_at - running_at])
                else:
                    stats['failed'] += 1

        summary_only = (
            # requested by user on command-line
            self.params.summary
            # automatically determined based on screen size
            or (len(rows) > capacity and self.params.verbose == 0)
        )
        if summary_only:
            # only print table with statistics
            table = PrettyTable(['state', 'num/tot', 'num/tot %'])
            table.header = False
            table.align['state'] = 'r'
            table.align['num/tot'] = 'c'
            table.align['num/tot %'] = 'r'

            for state, num in sorted(stats.items()):
                if (states is None) or (str(state) in states):
                    table.add_row([
                        state,
                        "%d/%d" % (num, tot),
                        "%.2f%%" % (100.0 * num / tot)
                    ])
        else:
            # print table of job status
            table = PrettyTable(["JobID", "Job name", "State", "Info"] + keys)
            table.align = 'l'
            for row in sorted(rows):
                table.add_row(row)
        print(table)

        if self.params.lifetimes is not None and len(lifetimes_rows) > 1:
            if self.params.lifetimes is sys.stdout:
                print("")
                print("Report on the job life times:")
                lifetimes_csv = csv.writer(
                    self.params.lifetimes, delimiter='\t')
            else:
                lifetimes_csv = csv.writer(self.params.lifetimes)
            lifetimes_csv.writerows(lifetimes_rows)

        # since `_get_tasks` swallows any exception raised by invalid
        # job IDs or corrupted files, let us determine the number of
        # failures by counting the number of times we actually run
        # this loop and then subtract from the number of times we
        # *should* have run, i.e., the number of arguments we were passed.
        failed = len(self.params.args) - tot
        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gget(GC3UtilsScript):
    """
Retrieve output files of a job.

Output files can only be retrieved once a job has reached the
'RUNNING' state; this command will print an error message if
no output files are available.

Output files can be retrieved multiple times until a job reaches
'TERMINATED' state: after that, the remote storage will be
released once the output files have been fetched.
    """
    def setup_options(self):
        self.add_param("-A",
                       action="store_true",
                       dest="all",
                       default=False,
                       help="Download *all* files of *all* tasks in a session."
                       " USE WITH CAUTION!")
        self.add_param("-d", "--download-dir",
                       action="store",
                       dest="download_dir",
                       default=None,
                       help="Destination directory (job id will be appended to"
                       " it); default is '.'")
        self.add_param("-f", "--overwrite",
                       action="store_true",
                       dest="overwrite",
                       default=False,
                       help="Overwrite files in destination directory")
        self.add_param("-c", "--changed-only",
                       action="store_true",
                       dest="changed_only",
                       default=False,
                       help="Only download files that were changed on remote"
                       " side.")

    def main(self):
        if self.params.all and len(self.params.args) > 0:
            raise gc3libs.exceptions.InvalidUsage(
                "Option '-A' conflicts with list of job IDs:"
                " use either '-A' or explicitly list task IDs.")

        if self.params.all:
            self.session = self._get_session(self.params.session)
            args = self._list_all_tasks()
            if len(args) == 0:
                self.log.info("No jobs in session: nothing to do.")
                return os.EX_OK
        else:
            args = self.params.args
            if len(args) == 0:
                self.log.error(
                    "No job IDs given on command line: nothing to do."
                    " Type '%s --help' for usage help."
                    # if we were called with an absolute path,
                    # presume the command has been found by the
                    # shell through PATH and just print the command name,
                    # otherwise print the exact path name.
                    % (os.path.basename(sys.argv[0])
                       if os.path.isabs(sys.argv[0])
                       else sys.argv[0]))
                return es.EX_USAGE
            self.session = self._get_session(self.params.session,
                                             task_ids=args)

        failed = 0
        download_dirs = set()
        for jobid in args:
            try:
                app = self.session.load(jobid)
                app.attach(self._core)

                if app.execution.state == Run.State.NEW:
                    raise gc3libs.exceptions.InvalidOperation(
                        "Job '%s' is not yet submitted. Output cannot be"
                        " retrieved" % app.persistent_id)
                elif app.execution.state == Run.State.TERMINATED:
                    raise gc3libs.exceptions.InvalidOperation(
                        "Output of '%s' already downloaded to '%s'"
                        % (app.persistent_id, app.output_dir))

                # XXX: this uses "private" code from `Application` and
                # `Core.fetch_output`
                app_download_dir = app._get_download_dir(
                    self.params.download_dir)
                # avoid downloading files twice for virtual tasks that
                # wrap an application (e.g., `RetryableTask`)
                if app_download_dir not in download_dirs:
                    self._core.fetch_output(
                        app,
                        download_dir=app_download_dir,
                        overwrite=self.params.overwrite,
                        changed_only=self.params.changed_only)
                    if app.changed:
                        self.session.store.replace(app.persistent_id, app)
                    # `fetch_output` is by default a no-op on tasks:
                    # try to detect this and skip all the messaging
                    # below (and also retry the directory, in case we
                    # have an Application with the same download
                    # destination)
                    if not os.path.exists(app_download_dir):
                        continue
                    # avoid downloading files twice -- see above
                    download_dirs.add(app_download_dir)
                else:
                    self.log.debug("Output directory '%s' already visited,"
                                   " not downloading again.", app_download_dir)
                # print message to user anyhow
                if app.execution.state == Run.State.TERMINATED:
                    print("Job final results were successfully retrieved in"
                          " '%s'" % (app_download_dir,))
                else:
                    print("A snapshot of job results was successfully"
                          " retrieved in '%s'" % (app_download_dir,))

            except Exception as ex:
                print("Failed retrieving results of job '%s': %s"
                      % (jobid, str(ex)))
                failed += 1
                continue

        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gkill(GC3UtilsScript):
    """
Cancel a submitted job.  Given a list of jobs, try to cancel each
one of them; exit with code 0 if all jobs were cancelled
successfully, and 1 if some job was not.

The command will print an error message if a job cannot be
canceled because it's in NEW or TERMINATED state, or if some other
error occurred.
    """

    def setup_options(self):
        self.add_param("-A", action="store_true", dest="all", default=False,
                       help="Remove all stored jobs. USE WITH CAUTION!")

    def main(self):
        if self.params.all and len(self.params.args) > 0:
            raise gc3libs.exceptions.InvalidUsage(
                "Option '-A' conflicts with list of job IDs to remove.")

        if self.params.all:
            self.session = self._get_session(self.params.session)
            args = self._list_all_tasks()
            if len(args) == 0:
                self.log.info("No jobs in session: nothing to do.")
                return os.EX_OK
        else:
            args = self.params.args
            if len(args) == 0:
                self.log.warning(
                    "No job IDs given on command line: nothing to do."
                    " Type '%s --help' for usage help."
                    # if we were called with an absolute path,
                    # presume the command has been found by the
                    # shell through PATH and just print the command name,
                    # otherwise print the exact path name.
                    % (os.path.basename(sys.argv[0])
                       if os.path.isabs(sys.argv[0])
                       else sys.argv[0]))
                return os.EX_USAGE
            self.session = self._get_session(self.params.session,
                                             task_ids=args)

        failed = 0
        for jobid in args:
            try:
                app = self.session.load(jobid)
                app.attach(self._core)

                self.log.debug("gkill: Job '%s' in state %s"
                               % (jobid, app.execution.state))
                if app.execution.state == Run.State.NEW:
                    raise gc3libs.exceptions.InvalidOperation(
                        "Job '%s' has never been submitted." % app)
                if app.execution.state == Run.State.TERMINATED:
                    raise gc3libs.exceptions.InvalidOperation(
                        "Job '%s' is already in terminal state" % app)
                else:
                    app.kill()
                    self.session.store.replace(jobid, app)

                    # or shall we simply return an ack message ?
                    print("Sent request to cancel job '%s'." % jobid)

            except Exception as ex:
                print("Failed canceling job '%s': %s" % (jobid, str(ex)))
                failed += 1
                continue

        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gtail(GC3UtilsScript):
    """
Display the last lines from a job's standard output or error stream.
Optionally, keep running and displaying the last part of the file
as more lines are written to the given stream.
    """
    def setup_args(self):
        """
        Override `GC3UtilsScript`:class: `setup_args` method since we
        don't operate on single jobs.
        """
        self.add_param('args',
                       nargs=1,
                       metavar='JOBID',
                       help="Job ID string identifying the single job to"
                       " operate upon.")

    def setup_options(self):
        self.add_param("-e", "--stderr",
                       action="store_true",
                       dest="stderr",
                       default=False,
                       help="show stderr of the job")
        self.add_param("-f", "--follow",
                       action="store_true",
                       dest="follow",
                       default=False,
                       help="output appended data as the file grows")
        self.add_param("-o", "--stdout",
                       action="store_true",
                       dest="stdout",
                       default=True,
                       help="show stdout of the job (default)")
        self.add_param("-n", "--lines",
                       dest="num_lines",
                       type=int,
                       default=10,
                       help="output the last N lines, instead of the last 10")

    def main(self):
        if len(self.params.args) == 0:
            self.log.error(
                "No job IDs given on command line: nothing to do."
                " Type '%s --help' for usage help."
                # if we were called with an absolute path,
                # presume the command has been found by the
                # shell through PATH and just print the command name,
                # otherwise print the exact path name.
                % (os.path.basename(sys.argv[0]) if os.path.isabs(sys.argv[0])
                   else sys.argv[0]))
            return os.EX_USAGE

        if self.params.stderr:
            stream = 'stderr'
        else:
            stream = 'stdout'

        self.session = self._get_session(self.params.session,
                                         task_ids=self.params.args)
        failed = 0
        for jobid in self.params.args:
            try:
                app = self.session.load(jobid)
                if app.execution.state == Run.State.UNKNOWN \
                        or app.execution.state == Run.State.SUBMITTED \
                        or app.execution.state == Run.State.NEW:
                    raise RuntimeError('Job output not yet available')
                app.attach(self._core)

                try:
                    if self.params.follow:
                        where = 0
                        while True:
                            file_handle = app.peek(stream)
                            self.log.debug("Seeking position %d in stream"
                                           " %s" % (where, stream))
                            file_handle.seek(where)
                            for line in file_handle.readlines():
                                print(line.strip())
                            where = file_handle.tell()
                            self.log.debug("Read up to position %d in stream"
                                           " %s" % (where, stream))
                            file_handle.close()
                            time.sleep(5)
                    else:
                        estimated_size = (
                            gc3libs.defaults.PEEK_FILE_SIZE * self.params.num_lines)
                        fh = app.peek(stream, offset=-estimated_size,
                                      size=estimated_size)
                        for line in fh.readlines()[-(self.params.num_lines):]:
                            print(line.strip())
                        fh.close()

                except gc3libs.exceptions.InvalidOperation:  # Cannot `peek()` on a task collection  # noqa
                    self.log.error(
                        "Task '%s' (of class '%s') has no defined output/error"
                        " streams. Ignoring.",
                        app.persistent_id,
                        app.__class__.__name__)
                    failed += 1
            except Exception as ex:
                print("Failed while reading content of %s for job '%s': %s"
                      % (stream, jobid, str(ex)))
                failed += 1

        # exit code is practically limited to 7 bits ...
        return min(failed, 126)


class cmd_gservers(GC3UtilsScript):
    """
List status of computational resources.
    """

    def setup_options(self):
        self.add_param("-n", "--no-update", action="store_false",
                       dest="update", default=True,
                       help="Do not update resource statuses;"
                       " only print what's in the local database.")
        self.add_param(
            "-p", "--print", action="store", dest="keys",
            metavar="LIST", default=None,
            help="Only print resource attributes whose name appears in"
            " this comma-separated list. (Attribute name is as given in"
            " the configuration file, or listed in the middle column"
            " in `gservers` output.)")

    def setup(self):
        """
        Override `GC3UtilsScript`:class: `setup` method since we
        don't need any `session` argument.
        """
        gc3libs.cmdline._Script.setup(self)

    def setup_args(self):
        """
        Override `GC3UtilsScript`:class: `setup_args` method since we
        don't operate on jobs but on resources.
        """
        self.add_param('args',
                       nargs='*',
                       metavar='RESOURCE',
                       help="Resource, string identifying the name of the"
                       " resources to check.")

    def main(self):
        if len(self.params.args) > 0:
            self._select_resources(* self.params.args)
            self.log.info(
                "Retained only resources: %s",
                str.join(",", [res['name']
                               for res in self._core.get_resources()]))

        if self.params.update:
            self._core.update_resources()

        resources = self._core.get_resources()

        for resource in sorted(resources, key=(lambda rsc: rsc.name)):
            if self.params.args and resource.name not in self.params.args:
                continue
            table = PrettyTable(['', resource.name, ' '])
            table.align = 'l'
            table.align[''] = 'r'

            # not all resources support the same keys...
            def output_if_exists(name, print_name):
                if hasattr(resource, name) and ((not self.params.keys) or
                                                name in self.params.keys):
                    table.add_row((name,
                                   ("( %s )" % print_name),
                                   getattr(resource, name)))
            output_if_exists('frontend', "Frontend host name")
            output_if_exists('type', "Access mode")
            output_if_exists('auth', "Authorization name")
            output_if_exists('updated', "Accessible?")
            output_if_exists('ncores', "Total number of cores")
            output_if_exists('queued', "Total queued jobs")
            output_if_exists('user_queued', "Own queued jobs")
            output_if_exists('user_run', "Own running jobs")
            # output_if_exists('free_slots', "Free job slots")
            output_if_exists('max_cores_per_job', "Max cores per job")
            output_if_exists('max_memory_per_core', "Max memory per core")
            output_if_exists('max_walltime', "Max walltime per job")
            output_if_exists('applications', "Supported applications")
            print(table)
            print('')


class cmd_gsession(GC3UtilsScript):
    """
`gsession` get info on a session.

Usage:

    gsession `command` [options] SESSION_DIR

commands are listed below, under `subcommands`.

To get detailed info on a specific command, run:

    gsession `command` --help
    """

    # Setup methods

    def _add_subcmd(self, name, func, help=None):
        subparser = self.subparsers.add_parser(name, help=help)
        subparser.set_defaults(func=func)
        subparser.add_argument('session')
        subparser.add_argument('-v', '--verbose', action='count')
        return subparser

    def setup(self):
        gc3libs.cmdline._Script.setup(self)

        self.subparsers = self.argparser.add_subparsers(
            title="subcommands",
            description="gsession accept the following subcommands. "
            "Each subcommand requires a `SESSION` directory as argument.")

        self._add_subcmd(
            'abort',
            self.abort_session,
            help="Kill all jobs related to a session and remove it from disk")
        self._add_subcmd(
            'delete',
            self.delete_session,
            help="Delete a session from disk.")
        subparser = self._add_subcmd(
            'list',
            self.list_jobs,
            help="Tree-like view of jobs related to a session.")
        subparser.add_argument('-r', '--recursive', action="store_true",
                               default=False,
                               help="Show all jobs contained in a task"
                               " collection, not only top-level jobs.")

        self._add_subcmd(
            'log',
            self.show_log,
            help="Show log entries for the session.")

    def setup_args(self):
        # prevent GC3UtilsScript.setup_args() to add the default JOBID
        # non optional argument
        pass

    def main(self):
        return self.params.func()

    # "working" methods
    def abort_session(self):
        """
        Called with subcommand `abort`.

        This method will open the desired session and will kill all
        the jobs which belongs to that session.

        Thiw method will return the number of jobs that have not been
        correctly aborted. If this number is greater than 125, then
        125 will be returned instead, since numbers above 125 have a
        special meaning for the shell.
        """
        self.session = self._get_session(self.params.session)

        rc = len(self.session.tasks)
        for task_id in list(self.session.tasks.keys()):
            task = self.session.tasks[task_id]
            task.attach(self._core)
            if task.execution.state == Run.State.TERMINATED:
                gc3libs.log.info("Not aborting '%s' which is already in"
                                 " TERMINATED state." % task)
                rc -= 1
                continue
            try:
                task.kill()
                task.free()
                rc -= 1
            except gc3libs.exceptions.Error as err:
                gc3libs.log.error(
                    "Could not abort task '%s': %s: %s",
                    task, err.__class__.__name__, err)

        if rc:
            gc3libs.log.error(
                "Not all tasks of the session have been aborted.")

        if rc > 125:
            # 126 and 127 error codes have special meanings.
            rc = 125
        self.session.save_all()
        return rc

    def delete_session(self):
        """
        Called with subcommand `delete`.

        This method will first call `abort` and then remove the
        current session.

        If the `abort` will fail or not all tasks are in TERMINATED
        state, it will not delete the session and will exit with the
        exit status of the `abort`.
        """
        try:
            rc = self.abort_session()
            if rc != 0:
                return rc
        except gc3libs.exceptions.InvalidArgument:
            raise RuntimeError('Session %s not found' % self.params.session)

        self.session.destroy()
        return 0

    def list_jobs(self):
        """
        Called with subcommand ``list``.

        List the content of a session, like ``gstat -n -v -s SESSION``
        does.  Unlike ``gstat``, though, display stops at the top-level
        jobs unless option `--recursive` is also given.

        With option ``--recursive``, indent job ids to show the
        tree-like organization of jobs in the task collections.
        """
        self.session = self._get_session(self.params.session)

        def print_app_table(app, indent, recursive):
            rows = []
            try:
                jobname = app.jobname
            except AttributeError:
                jobname = ''

            rows.append([indent + str(app.persistent_id),
                         jobname,
                         app.execution.state,
                         app.execution.info])
            if recursive and 'tasks' in app:
                indent = " "*len(indent) + '  '
                for task in app.tasks:
                    rows.extend(print_app_table(task, indent, recursive))
            return rows

        rows = []
        for app in list(self.session.tasks.values()):
            rows.extend(print_app_table(app, '', self.params.recursive))
        table = PrettyTable(["JobID", "Job name", "State", "Info"])
        table.align = 'l'
        for row in rows:
            table.add_row(row)
        print(table)

    def show_log(self):
        """
        Called when subcommand is `log`.

        This method will print the history of the jobs in SESSION in a
        logfile fashon
        """
        self.session = self._get_session(self.params.session)
        timestamps = []
        task_queue = list(self.session.tasks.values())
        while task_queue:
            app = task_queue.pop()
            for what, when, tags in app.execution.history._messages:
                timestamps.append((float(when), str(app), what))
            try:
                for child in app.tasks:
                    task_queue.append(child)
            except AttributeError:
                # Application class does not have a `tasks` attribute
                pass

        timestamps.sort(key=(lambda ts: ts[0]))
        for entry in timestamps:
            print("%s %s: %s" % (
                time.strftime(
                    "%b %d %H:%M:%S", time.localtime(entry[0])
                    ),
                str(entry[1]),
                entry[2]))


class cmd_gselect(GC3UtilsScript):
    """
Print IDs of jobs that match the specified criteria.
The criteria specified by command-line options will be
AND'ed together, i.e., a job must satisfy all of them
in order to be selected.
    """

    def setup_args(self):
        # No positional arguments allowed
        pass

    def setup_options(self):
        self.add_param(
            '--error-message', '--errmsg', metavar='REGEXP',
            help=("Select jobs such that a line in their error output (STDERR)"
                  " file matches the given regular expression pattern."),
        )
        self.add_param(
            '--input-file', metavar='FILENAME',
            help=("Select jobs with input file FILENAME"
                  " (only the file name is considered, not the full path)."),
            )
        self.add_param(
            '--jobname', '--job-name', metavar='REGEXP',
            dest='jobname', default='',
            help=("Select jobs whose name matches the supplied"
                  " regular expression (case insensitive).")
        )
        self.add_param(
            '--jobid', '--job-id', metavar='REGEXP',
            dest='jobid', default='',
            help=("Select jobs whose ID matches the supplied"
                  " regular expression (case insensitive).")
        )
        self.add_param(
            '-l', '--state', '--states',
            dest='states', default=None,
            help=("Select all jobs in one of the specified states"
                  " (comma-separated list). Valid states are: %s"
                  % str.join(", ", sorted(Run.State))),
            )
        self.add_param(
            '--output-file', metavar='FILENAME',
            help=("Select jobs with output file FILENAME"
                  " (only the file name is considered, not the full path)."),
            )
        self.add_param(
            '--output-message', '--outmsg', metavar='REGEXP',
            help=("Select jobs such that a line in their main output (STDOUT)"
                  " file matches the given regular expression pattern."),
        )
        self.add_param(
            '--successful', '--ok',
            action='store_true', dest='successful', default=False,
            help="Select jobs with non-zero exit code."
            )
        self.add_param(
            '--submitted-after', metavar='DATE',
            help="Select jobs submitted after the specified date.",
            )
        self.add_param(
            '--submitted-before', metavar='DATE',
            help="Select jobs submitted before the specified date.",
            )
        self.add_param(
            '--unsuccessful', '--failed',
            action='store_true', dest='unsuccessful', default=False,
            help="Select jobs with non-zero exit code."
            )

    def parse_args(self):
        self.criteria = []

        # --successful, --unsuccessful
        if self.params.successful and self.params.unsuccessful:
            raise gc3libs.exceptions.InvalidUsage(
                " Please use only one of the two options:"
                " `--successful` or `--unsuccessful`.")

        if self.params.successful:
            self.criteria.append(
                (self.params.successful, self.filter_successful, ()))

        if self.params.unsuccessful:
            self.criteria.append(
                (self.params.unsuccessful, self.filter_unsuccessful, ()))

        # --jobname, --job-name
        if self.params.jobname:
            try:
                self.jobname_re = re.compile(self.params.jobname, re.I)
                self.criteria.append(
                    (True, self.filter_by_jobname, (self.jobname_re,)))
            except re.error as err:
                raise gc3libs.exceptions.InvalidUsage(
                    "Regexp `%s` for option `--job-name` is invalid: %s"
                    % (self.params.jobname, err))

        # --jobid, --job-id
        if self.params.jobid:
            try:
                self.jobid_re = re.compile(self.params.jobid, re.I)
                self.criteria.append(
                    (True, self.filter_by_jobid, (self.jobid_re,)))

            except re.error as err:
                raise gc3libs.exceptions.InvalidUsage(
                    "Regexp `%s` for option `--job-id` is invalid: %s"
                    % (self.params.jobid, err))

        # --state
        if self.params.states is not None:
            self.allowed_states = set(i.upper()
                                      for i in self.params.states.split(','))
            invalid = self.allowed_states.difference(Run.State)
            if invalid:
                raise gc3libs.exceptions.InvalidUsage(
                    "Invalid state(s): %s" % str.join(", ", invalid))

            self.criteria.append(
                (True, self.filter_by_state, (self.allowed_states,)))

        # --submitted-after, --submitted-before
        self.submission_start = None

        # NOTE: short-cut `if self.params.submitted:` will *not* work here, as
        # the empty string is a valid value -- only `None` indicates that the
        # option was not given
        if self.params.submitted_after is not None:
            if self.params.submitted_after == '':
                gc3libs.log.warning(
                    "Empty date as argument of --submitted-after will be"
                    " interpreted as 'now'")

            try:
                self.submission_start = time.mktime(
                    Calendar().parse(self.params.submitted_after)[0])
            except Exception as ex:
                raise gc3libs.exceptions.InvalidUsage(
                    "Invalid value `%s` for --submitted-after argument: %s"
                    % (self.params.submitted_after, str(ex)))

        else:  # no `--submitted-after` option
            # a starting date is always needed; if user did not specify one,
            # then choose the beginning of time
            self.submission_start = 0.0

        # NOTE: short-cut `if self.params.submitted_before:` will *not* work
        # here, as the empty string is a valid value -- only `None` indicates
        # that the option was not given
        if self.params.submitted_before is not None:
            if self.params.submitted_before == '':
                gc3libs.log.warning(
                    "Empty date as argument of --submitted-before will be"
                    " interpreted as 'now'")
            try:
                self.submission_end = time.mktime(
                    Calendar().parse(self.params.submitted_before)[0])
            except Exception as err:
                raise gc3libs.exceptions.InvalidUsage(
                    "Invalid value `%s` for --submitted-before argument: %s"
                    % (self.params.submitted_before, err))

        else:  # no `--submitted-before` option
            # an ending date is always needed; if user did not specify one,
            # then choose the end of (UNIX) time
            self.submission_end = float(sys.maxsize)

        self.criteria.append(
            (True, self.filter_by_submission_date,
             (self.submission_start, self.submission_end)))

        # --input-file
        if self.params.input_file:
            self.criteria.append(
                (self.params.input_file or self.params.output_file,
                 self.filter_by_iofile,
                 (self.params.input_file, self.params.output_file)))

        # --error-message
        if self.params.error_message:
            self.criteria.append(
                (self.params.error_message,
                 self.filter_by_errmsg,
                 (self.params.error_message,)))

        # --output-message
        if self.params.output_message:
            self.criteria.append(
                (self.params.output_message,
                 self.filter_by_outmsg,
                 (self.params.output_message,)))

    def main(self):
        self.session = self._get_session(self.params.session)

        # Load tasks from the store
        current_jobs = list(self.session.iter_workflow())

        # pipeline of checks to perform; more expensive checks should come last
        # so they look at less jobs (do I long for LISP? Oh yes I do...)
        for cond, fn, args in self.criteria:
            if cond:
                current_jobs = fn(current_jobs, *args)
                if not current_jobs:
                    break

        # Print remaining job IDs, if any
        if current_jobs:
            print(str.join('\n',
                           (str(job.persistent_id) for job in current_jobs)))
        else:
            gc3libs.log.info("No jobs match the specified conditions.")

    @staticmethod
    def _filter_by_regexp(job_list, regexp, attribute):
        """
        Return list of items in `job_list` whose attribute
        named `attribute` matches the supplied (non-anchored) regexp.

        Jobs that do not possess an attribute with the specified name do *not*
        match.
        """
        # I know this could be written in one line but it would be
        # quite harder to read.
        matching_jobs = []
        for job in job_list:
            try:
                attrvalue = getattr(job, attribute)
            except AttributeError:
                continue
            if regexp.search(attrvalue):
                matching_jobs.append(job)
        return matching_jobs

    @staticmethod
    def filter_by_exitcode(job_list, codes):
        """
        Return list of tasks in `job_list` whose exit code
        is one of the given codes.

        If `codes` is empty, then return `job_list` unchanged.
        """
        if codes:
            return [job for job in job_list if job.execution.exitcode in codes]
        else:
            return job_list

    @staticmethod
    def filter_by_errmsg(job_list, msg):
        """
        Return list of tasks in `job_list` such that string `msg` occurs in the
        STDERR log file.
        """
        return [job for job in job_list
                if (hasattr(job, 'output_dir')
                    and utils.occurs(
                        msg, os.path.join(
                            job.output_dir,
                            (job.stdout if job.join else job.stderr)
                        )))]

    @staticmethod
    def filter_by_iofile(job_list, ifile=None, ofile=None):
        """
        Return list of items in `job_list` such that the `ifile` matches the
        *base name* of any input file, *and* `ofile` matches the base name of
        any output file.  If either one of `ifile` or `ofile` is ``None``, the
        corresponding check is turned off.
        """
        matching_jobs = []
        for job in job_list:
            # generic `Task` objects might not have `.inputs` or `.outputs`,
            # but they will be freely mixed with `Application` objects in a
            # session,xxxx so ignore errors here
            try:
                # `Application.inputs` is a `UrlKeyDict`
                inputs = [os.path.basename(url.path) for url in job.inputs]
            except AttributeError as err:
                gc3libs.log.debug(
                    "No input file data in task %s: %s."
                    " I'm turning off input file checks for this task",
                    job, err)
                inputs = []
            try:
                # `Application.inputs` is a `UrlValueDict`, so keys
                # are simple paths
                outputs = [os.path.basename(file) for file in job.outputs]
            except AttributeError as err:
                gc3libs.log.debug(
                    "No output file data in task %s: %s."
                    " I'm turning off output file checks for this task",
                    job, err)
                outputs = []
            matching = True
            if ifile and ifile not in inputs:
                matching = False
            if ofile and ofile not in outputs:
                matching = False
            if matching:
                matching_jobs.append(job)
        return matching_jobs

    @staticmethod
    def filter_by_jobid(job_list, regexp):
        """
        Return list of items in `job_list` whose ID (i.e., the `.persistent_id`
        attribute) matches the supplied regexp.  Jobs that do not have a
        `.persistent_id` attribute do *not* match.
        """
        return cmd_gselect._filter_by_regexp(job_list, regexp, 'persistent_id')

    @staticmethod
    def filter_by_jobname(job_list, regexp):
        """
        Return list of items in `job_list` whose `.jobname` attribute
        matches the supplied regexp.  Jobs that do not have a
        `jobname` attribute do *not* match.

        """
        return cmd_gselect._filter_by_regexp(job_list, regexp, 'jobname')

    @staticmethod
    def filter_by_outmsg(job_list, msg):
        """
        Return list of tasks in `job_list` such that string `msg` occurs in the
        STDOUT log file.
        """
        return [job for job in job_list
                if (hasattr(job, 'output_dir')
                    and utils.occurs(
                        msg, os.path.join(job.output_dir, job.stdout)
                    ))]

    @staticmethod
    def filter_successful(job_list):
        """
        Return list of tasks from `job_list` that were successfully terminated.
        """
        return [job for job in job_list
                if job.execution.returncode == 0]

    @staticmethod
    def filter_by_state(job_list, states):
        """
        Return list of elements in `job_list` whose state
        is one of the given states.
        """
        return [job for job in job_list if job.execution.state in states]

    @staticmethod
    def filter_unsuccessful(job_list):
        """
        Return list of tasks from `job_list` that were unsuccessfully
        terminated.
        """
        return [job for job in job_list
                if job.execution.returncode is not None
                and job.execution.returncode != 0]

    @staticmethod
    def filter_by_submission_date(job_list, start, end):
        """
        Return list of items in `job_list` which have been
        submitted within the range specified by `start` and `end`.
        """
        matching_jobs = []
        for job in job_list:
            submission_time = job.execution.timestamp.get(
                Run.State.SUBMITTED, None)
            if submission_time is None:
                # Jobs run by the ShellCmd backend transition directly to
                # RUNNING; use that timestamp if available.
                #
                # Since jobs in NEW state will not have any timestamp
                # at all, set the default to 0.0.
                submission_time = job.execution.timestamp.get(
                    Run.State.RUNNING, 0.0)

            if (submission_time <= end
                    and submission_time >= start):
                    matching_jobs.append(job)
        return matching_jobs


class cmd_gcloud(GC3UtilsScript):
    """
`gcloud` manage VMs created by the EC2 backend

Usage:

    gcloud `command` [options]

commands are listed below, under `subcommands`.

To get detailed info on a specific command, run:

    gcloud `command` --help
    """

    def _add_subcmd(self, name, func, help=None):
        subparser = self.subparsers.add_parser(name, help=help)
        subparser.set_defaults(func=func)
        subparser.add_argument('-v', '--verbose', action='count')
        return subparser

    def setup(self):
        gc3libs.cmdline._Script.setup(self)

        self.subparsers = self.argparser.add_subparsers(
            title="subcommands",
            description="gcloud accept the following subcommands.")

        listparser = self._add_subcmd(
            'list',
            self.list_vms,
            help='List VMs currently know to the Cloud backend.')
        listparser.add_argument(
            '-r', '--resource', metavar="NAME", dest="resource_name",
            default=None, help="Select resource by name.")
        listparser.add_argument(
            '-n', '--no-update', action="store_false", dest="update",
            help="Do not update job and flavor information;"
            " only print what is in the local database.")

        cleanparser = self._add_subcmd(
            'cleanup',
            self.cleanup_vms,
            help='Terminate VMs not currently running any job.')
        cleanparser.add_argument(
            '-r', '--resource', metavar="NAME", dest="resource_name",
            default=None, help="Limit to VMs running on this resource.")
        cleanparser.add_argument(
            '-n', '--dry-run',
            action="store_true", dest="dry_run", default=False,
            help="Do not perform any actual killing;"
            " just print what VMs would be terminated.")

        terminateparser = self._add_subcmd(
            'terminate',
            self.terminate_vm,
            help='Terminate a VM.')
        terminateparser.add_argument('ID', help='ID of the VM to terminate.',
                                     nargs='+')
        terminateparser.add_argument(
            '-r', '--resource', metavar="NAME", dest="resource_name",
            default=None, help="Select resource by name.")

        forgetparser = self._add_subcmd(
            'forget',
            self.forget_vm,
            help="Remove the VM from the list of known VMs, so that the EC2"
            " backend will stop submitting jobs on it. Please note that if you"
            " `forget` a VM, it will disappear from the output of"
            " `gcloud list`.")
        forgetparser.add_argument('ID',
                                  help='ID of the VM to "forget".', nargs='+')
        forgetparser.add_argument(
            '-r', '--resource', metavar="NAME", dest="resource_name",
            default=None, help="Select resource by name.")

        runparser = self._add_subcmd(
            'run',
            self.create_vm,
            help='Run a VM.')
        runparser.add_argument(
            '-r', '--resource', metavar="NAME", dest="resource_name",
            default=None, help="Select resource by name.")
        runparser.add_argument(
            '-f', '--flavor', dest="instance_type",
            default=None, help="Select instance flavor.")
        runparser.add_argument(
            '-i', '--image-id', metavar="ID", dest="image_id", default=None,
            help="Select the image id to use, if different from the one "
            "specified in the configuration file.")

    def setup_args(self):
        # prevent GC3UtilsScript.setup_args() to add the default JOBID
        # non optional argument
        pass

    def main(self):
        resources = [res for res in self._core.get_resources()
                     if res.type.startswith('ec2') or
                     res.type.startswith('openstack')]
        if self.params.resource_name:
            resources = [res for res in resources
                         if res.name == self.params.resource_name]
            if not resources:
                raise RuntimeError(
                    'No Cloud resource found matching name `%s`.'
                    '' % self.params.resource_name)

        if not resources:
            raise RuntimeError('No Cloud resource found.')

        self.resources = resources

        return self.params.func()

    @staticmethod
    def _print_vms(vms, res, header=True):
        table = PrettyTable()
        table.border = True
        if header:
            table.field_names = ["resource", "id", "state", "IP Address",
                                 "other IPs", "Nr. of jobs", "Nr. of cores",
                                 "image", "keypair"]

        images = []
        for vm in vms:
            nr_remote_jobs = 'N/A'
            ncores = 'N/A'
            status = ''
            image_name = ''
            ips = []
            if vm.id in res.subresources:
                if res.subresources[vm.id].updated:
                    nr_remote_jobs = str(res.subresources[vm.id].user_run)
                    ncores = str(res.subresources[vm.id].max_cores)

            if res.type.startswith('ec2'):
                ips = [vm.public_dns_name, vm.private_ip_address]
                status = vm.state
                image_name = vm.image_id
            elif res.type.startswith('openstack'):
                ips = vm.networks.get('private', []) + \
                    vm.networks.get('public', [])
                status = vm.status
                if not images:
                    images.extend(res._get_available_images())

                image_names = [x for x in images if x.id == vm.image['id']]
                if image_names:
                    image_name = image_names[0].name
                else:
                    image_name = 'UNKNOWN'
            if vm.preferred_ip in ips:
                ips.remove(vm.preferred_ip)

            table.add_row((res.name, vm.id, status, vm.preferred_ip,
                           str.join(', ', ips), nr_remote_jobs, ncores,
                           image_name, vm.key_name))

        print(table)

    @staticmethod
    def list_vms_in_resource(res, lock, update):
            printed = 0
            if update:
                res.get_resource_status()
            else:
                res._connect()

            vms = res._vmpool.get_all_vms()
            # draw title to separate output from different resources
            title = "VMs running on Cloud resource `%s` of type" \
                    " %s" % (res.name, res.type)
            separator = ('=' * len(title))

            # Acquire the lock before printing
            lock.acquire()
            gc3libs.log.debug(
                "list_vms: Acquiring lock for resource %s", res.name)
            print('')
            print(separator)
            print(title)
            print(separator)
            print('')

            # draw table of VMs running on resource `res`
            if vms:
                cmd_gcloud._print_vms(vms, res)
                printed += len(vms)

            if not printed:
                print("  no known VMs are currently running on this resource.")
            gc3libs.log.debug(
                "list_vms: Releasing lock for resource %s", res.name)
            lock.release()

    # Subcommand methods
    def list_vms(self):
        lock = mp.Lock()

        pool = []
        for res in self.resources:
            gc3libs.log.debug(
                "Creating a new process to get status of resource %s",
                res.name)

            def _run_list_vms():
                return self.list_vms_in_resource(res, lock, self.params.update)
            worker = mp.Process(target=_run_list_vms)
            pool.append(worker)
            worker.start()

        for worker in pool:
            worker.join()
        return 0

    def cleanup_vms(self):
        for res in self.resources:
            res.get_resource_status()

            # walk list of VMs and kill the unused ones
            vms = res._vmpool.get_all_vms()
            if vms:
                for vm in vms:
                    if res.subresources[vm.id].has_running_tasks():
                        if self.params.dry_run:
                            print("No job running on VM `%s` of resource `%s`;"
                                  " would terminate it." % (vm.id, res.name))
                        else:
                            # no dry run -- the real thing!
                            gc3libs.log.info(
                                "No job running on VM `%s` of resource `%s`;"
                                " terminating it ...", vm.id, res.name)
                            self._terminate_vm(vm.id, res)

        return 0

    def _find_resources_by_running_vm(self, vmid):
        """
        Returns a list of resources that are currently running a VM
        with id `vmid`
        """
        matching_res = []
        for resource in self.resources:
            if vmid in resource._vmpool:
                matching_res.append(resource)
        return matching_res

    def _terminate_vm(self, vmid, res=None):
        if res is None:
            matching_res = self._find_resources_by_running_vm(vmid)
        else:
            matching_res = [res]
        if len(matching_res) > 1:
            raise LookupError(
                "VM with ID `%s` have been found on multiple resources. Please"
                " specify the resource by running `gcloud terminate` with the"
                " `-r` option.")
        elif not matching_res:
            raise LookupError(
                "VM with id `%s` not found." % (vmid))

        resource = matching_res[0]
        vm = resource._vmpool.get_vm(vmid)
        gc3libs.log.info("Terminating VM `%s` on resource `%s`" %
                         (vmid, resource.name))
        if resource.type.startswith('ec2'):
            vm.terminate()
        elif resource.type.startswith('openstack'):
            vm.delete()
        else:
            gc3libs.log.error("Unsupported cloud provider %s." % resource.type)
        resource._vmpool.remove_vm(vmid)

    def terminate_vm(self):
        for resource in self.resources:
            resource.get_resource_status()
        errors = 0
        for vmid in self.params.ID:
            try:
                self._terminate_vm(vmid)
            except LookupError as ex:
                gc3libs.log.warning(str(ex))
                errors += 1
        return errors

    def _forget_vm(self, vmid):
        matching_res = self._find_resources_by_running_vm(vmid)
        if len(matching_res) > 1:
            raise LookupError("VM with ID `%s` have been found on multiple "
                              "resources. Please specify the resource by "
                              "running `grun` with the `-r` option.")
        elif not matching_res:
            raise LookupError(
                "VM with id `%s` not found." % (vmid))

        resource = matching_res[0]
        resource._vmpool.remove_vm(vmid)

    def forget_vm(self):
        for resource in self.resources:
            # calling `_connect()` will load the list of available VMs
            resource._connect()
        errors = 0
        for vmid in self.params.ID:
            try:
                self._forget_vm(vmid)
            except LookupError as ex:
                gc3libs.log.warning(str(ex))
                errors += 1
        return errors

    def create_vm(self):
        if len(self.resources) > 1:
            raise RuntimeError("Please specify the resource where you want to "
                               "create the VM by supplying the `-r` option.")

        resource = self.resources[0]
        resource._connect()
        image_id = self.params.image_id or resource.image_id
        instance_type = self.params.instance_type or resource.instance_type
        vm = resource._create_instance(image_id, instance_type=instance_type)
        resource._vmpool.add_vm(vm)
        self._print_vms([vm], resource)



#! /usr/bin/env python

"""
Utility classes and methods for dealing with URLs.
"""

# Copyright (C) 2011-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import str


from collections import namedtuple
import os
import urllib.parse

from .persistence.serialization import Persistable
from .utils import to_str

__docformat__ = 'reStructuredText'


_UrlFields = namedtuple('_UrlFields', [
    # watch out that this is *exactly* the order used in
    # `_UrlFields.__new__()` calls below!
    'scheme', 'netloc', 'path',
    'hostname', 'port', 'query',
    'username', 'password', 'fragment'
])


class Url(_UrlFields):

    """
    Represent a URL as a named-tuple object.  This is an immutable
    object that cannot be changed after creation.

    The following read-only attributes are defined on objects of class `Url`.

    =========   =====   ===================================  ==================
    Attribute   Index   Value                                if not present
    =========   =====   ===================================  ==================
    scheme      0       URL scheme specifier                 empty string
    netloc      1       Network location part                empty string
    path        2       Hierarchical path                    empty string
    query       3       Query component                      empty string
    hostname    4       Host name (lower case)               None
    port        5       Port number as integer (if present)  None
    username    6       User name                            None
    password    7       Password                             None
    fragment    8       URL fragment (part after ``#``)      empty string
    =========   =====   ===================================  ==================

    There are two ways of constructing `Url` objects:

    * By passing a string `urlstring`::

        >>> u = Url('http://www.example.org/data')

        >>> u.scheme == 'http'
        True
        >>> u.netloc == 'www.example.org'
        True
        >>> u.path == '/data'
        True

      The default URL scheme is ``file``::

        >>> u = Url('/tmp/foo')
        >>> u.scheme == 'file'
        True
        >>> u.path == '/tmp/foo'
        True

      However, if a ``#`` character is present in the path name, it
      will be taken as separating the path from the "fragment"::

        >>> u = Url('/tmp/foo#1')
        >>> u.path == '/tmp/foo'
        True
        >>> u.fragment == '1'
        True

      Please note that extra leading slashes '/' are interpreted as
      the begining of a network location:

        >>> u = Url('//foo/bar')
        >>> u.path == '/bar'
        True
        >>> u.netloc == 'foo'
        True
        >>> Url('///foo/bar').path == '/foo/bar'
        True

      (Check RFC 3986 http://tools.ietf.org/html/rfc3986)

      If `force_abs` is `True` (default), then the `path`
      attribute is made absolute, by calling `os.path.abspath` if
      necessary::

        >>> u = Url('foo/bar', force_abs=True)
        >>> os.path.isabs(u.path)
        True

      Otherwise, if `force_abs` is `False`, then the `path`
      attribute stores the passed string unchanged::

        >>> u = Url('foo', force_abs=False)
        >>> os.path.isabs(u.path)
        False
        >>> u.path == 'foo'
        True

      Other keyword arguments can specify defaults for missing parts
      of the URL::

        >>> u = Url('/tmp/foo', scheme='file', netloc='localhost')
        >>> u.scheme == 'file'
        True
        >>> u.netloc == 'localhost'
        True
        >>> u.path == '/tmp/foo'
        True

      Query attributes are also supported::

        >>> u = Url('http://www.example.org?foo=bar')
        >>> u.query == 'foo=bar'
        True

      and so are fragments::

        >>> u = Url('postgresql://user@db.example.org#table=data')
        >>> u.fragment == 'table=data'
        True

    * By passing keyword arguments only, to construct an `Url` object
      with exactly those values for the named fields::

        >>> u = Url(scheme='http', netloc='www.example.org', path='/data')

      In this form, the `force_abs` parameter is ignored.

    See also: http://goo.gl/9WcRvR
    """
    __slots__ = ()

    def __new__(cls, urlstring=None, force_abs=True,
                scheme='file', netloc='', path='',
                hostname=None, port=None, query='',
                username=None, password=None, fragment=''):
        """
        Create a new `Url` object.  See the `Url`:class: documentation
        for invocation syntax.
        """
        if urlstring is not None:
            if isinstance(urlstring, Url):
                # copy constructor
                return _UrlFields.__new__(cls,
                    urlstring.scheme, urlstring.netloc, urlstring.path,
                    urlstring.hostname, urlstring.port, urlstring.query,
                    urlstring.username, urlstring.password, urlstring.fragment
                )
            else:
                # XXX: `future` provides a backport of Py3's
                # `urlsplit()` function; however, the implementation
                # requires that all arguments have the same string
                # type of the 1st one (i.e., `urlstring` here).  This
                # is a source of problems as default value for
                # `scheme` is a load-time constant (and thus either
                # unicode or byte string depending on Python version
                # and on whether `unicode_literals` is in effect), but
                # `urlstring`'s type depends on how it was produced
                # (user input, pickled file, result of some library
                # call, etc). So we convert `urlstring` to `future`'s
                # `newstr` type to match the value of `scheme` set at
                # load-time.
                urlstring = to_str(urlstring, 'filesystem')

                # parse `urlstring` and use kwd arguments as default values
                try:
                    urldata = urllib.parse.urlsplit(
                        urlstring,
                        scheme=to_str(scheme, 'filesystem'),
                        allow_fragments=True)
                    if urldata.path is not None:
                        path = urldata.path
                    if urldata.scheme == 'file' \
                       and not os.path.isabs(path) \
                       and force_abs:
                        path = os.path.abspath(path)
                    # Python 2.6 parses fragments only for http(s),
                    # for any other scheme, the fragment is returned as
                    # part of the path...
                    if '#' in path:
                        path_, fragment_ = path.split('#')
                        urldata = urllib.parse.SplitResult(
                            urldata.scheme, urldata.netloc,
                            path_, urldata.query, fragment_)
                    return _UrlFields.__new__(cls,
                        urldata.scheme or to_str(scheme, 'filesystem'),
                        urldata.netloc or to_str(netloc, 'filesystem'),
                        path,
                        urldata.hostname or to_str(hostname, 'filesystem'),
                        urldata.port or to_str(port, 'filesystem'),
                        urldata.query or to_str(query, 'filesystem'),
                        urldata.username or to_str(username, 'filesystem'),
                        urldata.password or to_str(password, 'filesystem'),
                        urldata.fragment or to_str(fragment, 'filesystem'),
                    )
                except (ValueError, TypeError, AttributeError) as err:
                    raise ValueError(
                        "Cannot parse string '%s' as a URL: %s: %s"
                        % (urlstring, err.__class__.__name__, err))
        else:
            # no `urlstring`, use kwd arguments
            return _UrlFields.__new__(
                cls,
                to_str(scheme, 'filesystem'),
                to_str(netloc, 'filesystem'),
                to_str(path, 'filesystem'),
                to_str(hostname, 'filesystem'),
                to_str(port, 'filesystem'),
                to_str(query, 'filesystem'),
                to_str(username, 'filesystem'),
                to_str(password, 'filesystem'),
                to_str(fragment, 'filesystem')
            )

    def __getnewargs__(self):
        """Support pickling/unpickling `Url` class objects."""
        return (None, False,  # urlstring, force_abs
                self.scheme, self.netloc, self.path, self.hostname, self.port,
                self.query, self.username, self.password)

    # In Py3, for some reason `Url` does not inherit `_UrlFields`'
    # definition of `__hash__()`, thus leading to a "Unhashable type:
    # Url" errors.  So let's be explicit and state that a `Url` is
    # nothing but a `_UrlField` with some convenience methods added.
    def __hash__(self):
        return _UrlFields.__hash__(self)

    def __repr__(self):
        """
        Return a printed representation of this object, such that
        `eval(repr(x)) == x`.
        """
        return (
            "Url(scheme=%r, netloc=%r, path=%r, hostname=%r,"
            " port=%r, query=%r, username=%r, password=%r, fragment=%r)"
            %
            (self.scheme,
             self.netloc,
             self.path,
             self.hostname,
             self.port,
             self.query,
             self.username,
             self.password,
             self.fragment))

    def __str__(self):
        """
        Return a URL string corresponding to the URL object `urldata`.

            >>> u = Url('gsiftp://gridftp.example.org:2811/data')
            >>> str(u)
            'gsiftp://gridftp.example.org:2811/data'
            >>> u = Url('swift://swift.example.org:8080/v1?querystring')
            >>> str(u)
            'swift://swift.example.org:8080/v1?querystring'

        If the URL was constructed by parsing a URL string, a
        different string can result, although equivalent according to
        the URL definition::

            >>> u1 = Url('/tmp/foo')
            >>> str(u1)
            'file:///tmp/foo'

        """
        url = self.path
        # XXX: assumes all our URLs are of netloc-type!
        if url and not url.startswith('/'):
            url = '/' + url
        url = '//' + (self.netloc or '') + url
        if self.scheme:
            url = self.scheme + ':' + url
        if self.query:
            url += '?%s' % self.query
        if self.fragment:
            url += '#%s' % self.fragment
        return url

    def __eq__(self, other):
        """
        Return `True` if `self` and `other` are equal `Url` objects,
        or if their string representations match.

        Examples:

          >>> u = Url('/tmp/foo')
          >>> u == Url('/tmp/foo')
          True
          >>> u == Url('file:///tmp/foo')
          True
          >>> u == Url('http://example.org/')
          False

          >>> u == str(u)
          True
          >>> u == '/tmp/foo'
          True
          >>> u == 'file:///tmp/foo'
          True
          >>> u == 'http://example.org'
          False

          >>> u == 42
          False

          >>> u == Url('file:///tmp/foo?bar')
          False

        """
        try:
            # The `tuple.__eq__` call can only be used if both `self`
            # and `other` are `tuple` subclasses; we know that `self`
            # is, but we need to check `other`.
            return ((isinstance(other, tuple) and tuple.__eq__(self, other))
                    or str(self) == str(other)
                    or tuple.__eq__(self, Url(other)))
        except ValueError:
            # `other` is not a URL and cannot be made into one
            return False

    def __ne__(self, other):
        """
        The opposite of `__eq__`.
        """
        return not self.__eq__(other)

    def adjoin(self, relpath):
        """
        Return a new `Url`, constructed by appending `relpath` to the
        path section of this URL.

        Example::

            >>> u0 = Url('http://www.example.org')
            >>> u1 = u0.adjoin('data')
            >>> str(u1)
            'http://www.example.org/data'

            >>> u2 = u1.adjoin('moredata')
            >>> str(u2)
            'http://www.example.org/data/moredata'

        Even if `relpath` starts with `/`, it is still appended to the
        path in the base URL::

            >>> u3 = u2.adjoin('/evenmore')
            >>> str(u3)
            'http://www.example.org/data/moredata/evenmore'

        Optional query attribute is left untouched::

            >>> u4 = Url('http://www.example.org?bar')
            >>> u5 = u4.adjoin('foo')
            >>> str(u5)
            'http://www.example.org/foo?bar'

        """
        if relpath.startswith('/'):
            relpath = relpath[1:]
        return Url(scheme=self.scheme, netloc=self.netloc,
                   path=os.path.join((self.path or '/'), relpath),
                   hostname=self.hostname, port=self.port,
                   username=self.username, password=self.password,
                   query=self.query, fragment=self.fragment)


class _UrlDict(dict):
    """
    Base class for `UrlKeyDict`:class: and `UrlValueDict`:class:
    """
    __slots__ = (
        '_force_abs',
    )

    def __init__(self, iter_or_dict=None, force_abs=False, **extra_kv):
        self._force_abs = force_abs
        if iter_or_dict is not None:
            try:
                # if `iter_or_dict` is a dict-like object, then it has
                # `iteritems()`
                for k, v in iter_or_dict.items():
                    self[k] = v
            except AttributeError:
                # then assume `iter_or_dict` is an iterator over (key, value)
                # pairs
                for k, v in iter_or_dict:
                    self[k] = v
        if extra_kv:
            for k, v in extra_kv.items():
                self[k] = v

    def __repr__(self):
        return ('{0}({1}, force_abs={2})'
                .format(self.__class__.__name__, dict(self), self._force_abs))



class UrlKeyDict(_UrlDict):

    """
    A dictionary class enforcing that all keys are URLs.

    Strings and/or objects returned by `urlparse` can
    be used as keys.  Setting a string key automatically
    translates it to a URL:

       >>> d = UrlKeyDict()
       >>> d['/tmp/foo'] = 1
       >>> for k in d.keys(): print (type(k), k.path) # doctest:+ELLIPSIS
       <class '....Url'> /tmp/foo

    Retrieving the value associated with a key works with both the
    string or the url value of the key:

        >>> d['/tmp/foo']
        1
        >>> d[Url('/tmp/foo')]
        1

    Key lookup can use both the string or the `Url` value as well:

        >>> '/tmp/foo' in d
        True
        >>> Url('/tmp/foo') in d
        True
        >>> 'file:///tmp/foo' in d
        True
        >>> 'http://example.org' in d
        False

    Class `UrlKeyDict` supports initialization by copying items from
    another `dict` instance or from an iterable of (key, value)
    pairs::

        >>> d1 = UrlKeyDict({ '/tmp/foo':'foo', '/tmp/bar':'bar' })
        >>> d2 = UrlKeyDict([ ('/tmp/foo', 'foo'), ('/tmp/bar', 'bar') ])
        >>> d1 == d2
        True

    An empty `UrlKeyDict` instance is returned by the constructor
    when called with no parameters::

        >>> d0 = UrlKeyDict()
        >>> len(d0)
        0

    If `force_abs` is `True`, then all paths are converted to
    absolute ones in the dictionary keys.

        >>> d = UrlKeyDict(force_abs=True)
        >>> d['foo'] = 1
        >>> for k in d.keys(): print(os.path.isabs(k.path))
        True

        >>> d = UrlKeyDict(force_abs=False)
        >>> d['foo'] = 2
        >>> for k in d.keys(): print(os.path.isabs(k.path))
        False

    """

    __slots__ = ()

    def __setitem__(self, key, value):
        if not isinstance(key, Url):
            key = Url(key, self._force_abs)
        super(UrlKeyDict, self).__setitem__(key, value)

    # these two methods are necessary to have key-lookup work with
    # strings as well

    def __getitem__(self, key):
        if not isinstance(key, Url):
            key = Url(key, self._force_abs)
        return super(UrlKeyDict, self).__getitem__(key)

    def __contains__(self, key):
        if not isinstance(key, Url):
            key = Url(key, self._force_abs)
        return super(UrlKeyDict, self).__contains__(key)


class UrlValueDict(_UrlDict):

    """
    A dictionary class enforcing that all values are URLs.

    Strings and/or objects returned by `urlparse` can
    be used as values.  Setting a string value automatically
    translates it to a URL:

       >>> d = UrlValueDict()
       >>> d[1] = '/tmp/foo'
       >>> d[2] = Url('file:///tmp/bar')
       >>> for v in d.values(): print (type(v), v.path) # doctest:+ELLIPSIS
       <class '....Url'> /tmp/foo
       <class '....Url'> /tmp/bar

    Retrieving the value associated with a key always returns the
    URL-type value, regardless of how it was set::

        >>> d[1] == Url(scheme='file', netloc='', path='/tmp/foo', \
              hostname=None, port=None, query='', \
              username=None, password=None, fragment='')
        True

    Class `UrlValueDict` supports initialization by any of the
    methods that work with a plain `dict` instance::

        >>> d1 = UrlValueDict({ 'foo':'/tmp/foo', 'bar':'/tmp/bar' })
        >>> d2 = UrlValueDict([ ('foo', '/tmp/foo'), ('bar', '/tmp/bar') ])
        >>> d3 = UrlValueDict(foo='/tmp/foo', bar='/tmp/bar')

        >>> d1 == d2
        True
        >>> d2 == d3
        True

    In particular, an empty `UrlDict` instance is returned by the
    constructor when called with no parameters::

        >>> d0 = UrlValueDict()
        >>> len(d0)
        0

    If `force_abs` is `True`, then all paths are converted to
    absolute ones in the dictionary values.

        >>> d = UrlValueDict(force_abs=True)
        >>> d[1] = 'foo'
        >>> for v in d.values(): print(os.path.isabs(v.path))
        True

        >>> d = UrlValueDict(force_abs=False)
        >>> d[2] = 'foo'
        >>> for v in d.values(): print(os.path.isabs(v.path))
        False

    """

    __slots__ = ()

    def __setitem__(self, key, value):
        if not isinstance(value, Url):
            value = Url(value, self._force_abs)
        super(UrlValueDict, self).__setitem__(key, value)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="url",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python

"""
export SUBJECTS_DIR=/our/out/data/path

 ************
 DATA INPUT
 ************
recon-all -i in_data_path/s01/TP1/T1w_b.nii.gz -subjid s01.cross.TP1

 ************
 CROSS SECTIONAL PROCESSING s01
 ************
recon-all -s s01.cross.TP1 -all

# Optional

 ************
 LONG PROCESSING s01
 ************
recon-all -long s01.cross.TP1 s01.base -all -qcache
"""

from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import subprocess
import argparse

NII_EXTENSION = "nii"
FSAVERAGE = "fsaverage"
FS_SUBJECT_FSAVERAGE = os.path.join(os.environ.get("FREESURFER_HOME", ''),"subjects",FSAVERAGE)
FS_SEQ_LONG = "long"
FS_SEQ_CROSS = "cross"
FS_SEQ_DEFAULT = FS_SEQ_CROSS
FS_SEQ=[FS_SEQ_LONG,FS_SEQ_CROSS]

# def Usage():
#     print ("Usage: gnift_wrapper.py [OPTIONS] <subject name> <NIFIT input file> <subjects dir>")

def RunFreesurfer():
    """
    By default the input files are in the same local directory as the wrapper executed.
    """

    parser = argparse.ArgumentParser(description='Run Freesurfer.')
    parser.add_argument('subject', metavar='SUBJECT',
                        help='Subject name')
    parser.add_argument('nifti', metavar='NIFTI',
                        help='NIFTI input file')
    parser.add_argument('output', metavar='OUTPUT',
                        help='Subject dir')
    parser.add_argument('--seq', dest='sequence', action="append",
                        default=[FS_SEQ_CROSS],
                        help='Freesurfer sequence. Valid values %s' % FS_SEQ)

    args = parser.parse_args()

    # Create output folder and add simlink to FSAVERAGE
    os.mkdir(args.output)
    os.symlink(FS_SUBJECT_FSAVERAGE,os.path.join(args.output,FSAVERAGE))

    # Verisfy input arguments
    try:
        assert os.path.isfile(args.nifti), \
            "Input NIFTI file %s not found" % args.nifti

        if args.sequence:
            args.sequence = FS_SEQ_DEFAULT
    except AssertionError as ex:
        raise OSError(ex.message)

    input_nii = dict()
    cross_files = []
    os.environ["SUBJECTS_DIR"] = args.output
    quality_checker = os.environ["QA_TOOLS"]

    if FS_SEQ_CROSS in args.sequence:

        # DATA INPUT and  CROSS SECTIONAL PROCESSING
        print("Start DATA INPUT on %s" % args.subject)
        "Example: recon-all -i file.nii.gz -subjid s01.cross.TP1"
        command="recon-all -i %s -subjid %s.crossTP1" % (args.nifti,args.nifti.split(".")[0])
        runme(command)

        print("Start CROSS SECTIONAL PROCESSING")
        "Example: recon-all -s s01.cross.TP1 -all"
        command="recon-all -s %s.crossTP1 -all" % args.nifti.split(".")[0]
        runme(command)

    if FS_SEQ_LONG in args.sequence:
        # LONG PROCESSING s01
        print("Start LONG PROCESSING on %s" % args.subject)
        # Example: recon-all -long s01.cross.TP1 s01.base -all -qcache
        command = "recon-all -long %scross -all -qcache" % args.subject
        runme(command)

def runme(command):
    """
    Comodity function to run commands using `subprocess` module
    Input: command to run
    Output: none
    Raise Exception in case command fails
    """
    proc = subprocess.Popen(
        [command],
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE)

    print("Running command %s" % command)
    (stdout, stderr) = proc.communicate()

    if proc.returncode != 0:
        print("Execution failed with exit code: %d" % proc.returncode)
        print("Output message: %s" % stdout)
        print("Error message: %s" % stderr)
        raise Exeption(stderr)

if __name__ == '__main__':
    sys.exit(RunFreesurfer())



#! /usr/bin/env python

"""
Collection of tools to supplement optimization algorithm
:class:`optimizer.EvolutionaryAlgorithm <gc3libs.optimizer.EvolutionaryAlgorithm>`.

Include a list of desired tools in param `after_update_opt_state` of
:class:`optimizer.EvolutionaryAlgorithm <gc3libs.optimizer.EvolutionaryAlgorithm>`.
"""

# Copyright (C) 2011, 2012, 2013  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
__docformat__ = 'reStructuredText'


import os
import sys
import logging

log = logging.getLogger("gc3.gc3libs")


class plot_population(object):

    '''Plot the 2-dimensional population of an
       :class:`gc3libs.optimizer.EvolutionaryAlgorithm` instance.
       If the population is not 2-d an error message appears and no
       plot is created.

       :param str figure_dir: Path to the directory where plots should be stored.
                              Directory will be created if non-existent.
    '''

    def __init__(self, figure_dir):
        if not os.path.isdir(figure_dir):
            os.mkdir(figure_dir)
        self.figure_dir = figure_dir

    def __call__(self, algo):
        '''Plot population for `algo`.

           :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`.
        '''
        log.debug('entering plot_population.__call__')
        pop = algo.pop
        if not algo.dim == 2:
            algo.logger.critical(
                'plot_population is implemented only for algo.dim = 2')
            return
        import matplotlib
        matplotlib.use('SVG')
        import matplotlib.pyplot as plt
        x = pop[:, 0]
        y = pop[:, 1]
        # determine bounds
        x0_min = algo.pop[:, 0].min()
        x1_min = algo.pop[:, 1].min()
        x0_max = algo.pop[:, 0].max()
        x1_max = algo.pop[:, 1].max()

        # Determine plotting area
        x_dif = x0_max - x0_min
        y_dif = x1_max - x1_min
        scaleFac = 0.3
        x0_min_plot = x0_min - scaleFac * x_dif
        x0_max_plot = x0_max + scaleFac * x_dif
        x1_min_plot = x1_min - scaleFac * y_dif
        x1_max_plot = x1_max + scaleFac * y_dif

        # make plot
        fig = plt.figure()
        ax = fig.add_subplot(111)

        ax.scatter(x, y)
        # x box constraints
        ax.plot([x0_min, x0_min], [x1_min_plot, x1_max_plot])
        ax.plot([x0_max, x0_max], [x1_min_plot, x1_max_plot])
        ax.plot([x0_min_plot, x0_max_plot], [x1_min, x1_min])
        ax.plot([x0_min_plot, x0_max_plot], [x1_max, x1_max])

        # all other linear constraints
        # c_x_min_plot = algo.filter_fn.linearConstr(x_min_plot)
        # c_xmax = algo.filter_fn.linearConstr(xmax)
        # for ixC in range(len(c_x_min_plot)):
        # ax.plot([x_min_plot, xmax], [c_x_min_plot[ixC], c_xmax[ixC]])
        ax.axis(xmin=x0_min_plot, xmax=x0_max_plot,
                ymin=x1_min_plot, ymax=x1_max_plot)

        ax.set_xlabel('x')
        ax.set_ylabel('y')

        ax.set_title('Best: x %s, f(x) %f' % (algo.best_x, algo.best_y))

        fig.savefig(os.path.join(self.figure_dir, 'pop%d' % (algo.cur_iter)))


def print_stats(algo, output=sys.stdout):
    '''Print summary statistics for `algo`.

       :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`.
       :param output: Output stream.
    '''
    output.write(
        'Iteration: %d,  x: %s f(x): %f\n' %
        (algo.cur_iter, algo.best_x, algo.best_y))


def log_stats(algo, logger=logging.getLogger()):
    '''Log summary statistics for `algo`.

       :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`.
    '''
    logger.info('Iteration: %d,  x: %s f(x): %f',
                algo.cur_iter, algo.best_x, algo.best_y)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011-2014, 2016  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

import datetime
import os
import shutil
import tempfile

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.backends.lsf import LsfLrms
from gc3libs.quantity import Duration, hours, Memory, GB

_datetime_date = None

files_to_remove = []


def setUpModule():
    """Mock the `datetime.date.today()` outcome in order to make the LSF
    parsing independent from the testing date."""
    # save the original `datetime.date` to restore it in `tearDownModule()`
    import datetime
    global _datetime_date
    _datetime_date = datetime.date
    # mock features of `datetime.date.today()` that are actually used
    # in `LsfLrms._parse_date()`

    class MockDate(object):

        def __init__(self, real):
            self.__date = real

        def __getattr__(self, name):
            return getattr(self.__date, name)

        def __call__(self, *args, **kwargs):
            return self.__date(*args, **kwargs)
    datetime.date = MockDate(datetime.date)

    class Today(object):

        def __init__(self):
            self.year = 2012
            self.month = 12
    datetime.date.today = Today


def tearDownModule():
    # restore the original `datetime.date`
    global _datetime_date
    import datetime
    datetime.date = _datetime_date
    for fname in files_to_remove:
        if os.path.isdir(fname):
            shutil.rmtree(fname)
        else:
            os.remove(fname)


def test_get_command():
    (fd, tmpfile) = tempfile.mkstemp()
    files_to_remove.append(tmpfile)
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=lsf
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
bsub = /usr/local/bin/bsub -R lustre
bjobs = /usr/local/bin/bjobs
lshosts = /usr/local/sbin/lshosts # comments are ignored!

lsf_continuation_line_prefix_length = 12
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    b = cfg.make_resources()['example']

    assert b.bsub == ['/usr/local/bin/bsub', '-R', 'lustre']

    assert b._bjobs == '/usr/local/bin/bjobs'
    assert b._lshosts == '/usr/local/sbin/lshosts'

    assert b._CONTINUATION_LINE_START == 12 * ' '


def test_bjobs_output_done1():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    bjobs_output = """
Job <131851>, Job Name <ChromaExtractShort>, User <wwolski>, Project <default>,
                     Status <DONE>, Queue <pub.8h>, Job Priority <50>, Command
                     <ChromatogramExtractor -in /cluster/scratch/malars/openswa
                     th/data/AQUA_fixed_water/split_napedro_L120224_001_SW-400A
                     QUA_no_background_2ul_dilution_10/split_napedro_L120224_00
                     1_SW-400AQUA_no_background_2ul_dilution_10_28.mzML.gz -tr
                     /cluster/scratch/malars/openswath/assays/iRT/DIA_iRT.TraML
                      -out split_napedro_L120224_001_SW-400AQUA_no_background_2
                     ul_dilution_10_28._rtnorm.chrom.mzML -is_swath -min_upper_
                     edge_dist 1 -threads 2>, Share group charged </lsf_biol_al
                     l/lsf_biol_other/wwolski>
Tue Jul 24 10:03:15: Submitted from host <brutus3>, CWD <$HOME/.gc3pie_jobs/lrm
                     s_job.YNZmU17755/.>, Output File <lsf.o%J>, Requested Reso
                     urces <select[mem<70000 && lustre] order[-ut] rusage[mem=1
                     000,m=1]>, Login Shell </bin/sh>, Specified Hosts <thin+9>
                     , <single+8>, <smp16+6>, <smp24+5>, <smp48+4>;

 RUNLIMIT
 480.0 min of a6122
Tue Jul 24 10:04:19: Started on <a6122>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
                     obs/lrms_job.YNZmU17755/.>;
Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.

 MEMORY USAGE:
 MAX MEM: 41 Mbytes;  AVG MEM: 41 Mbytes

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -  1000M     -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched 4000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
"""
    jobstatus = lsf._parse_stat_output(bjobs_output, '')
    assert jobstatus.state == gc3libs.Run.State.TERMINATING
    assert jobstatus.termstatus == (0, 0)


def test_bjobs_output_for_accounting():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    bjobs_output = """
Job <131851>, Job Name <ChromaExtractShort>, User <wwolski>, Project <default>,
                     Status <DONE>, Queue <pub.8h>, Job Priority <50>, Command
                     <ChromatogramExtractor -in /cluster/scratch/malars/openswa
                     th/data/AQUA_fixed_water/split_napedro_L120224_001_SW-400A
                     QUA_no_background_2ul_dilution_10/split_napedro_L120224_00
                     1_SW-400AQUA_no_background_2ul_dilution_10_28.mzML.gz -tr
                     /cluster/scratch/malars/openswath/assays/iRT/DIA_iRT.TraML
                      -out split_napedro_L120224_001_SW-400AQUA_no_background_2
                     ul_dilution_10_28._rtnorm.chrom.mzML -is_swath -min_upper_
                     edge_dist 1 -threads 2>, Share group charged </lsf_biol_al
                     l/lsf_biol_other/wwolski>
Tue Jul 24 10:03:15: Submitted from host <brutus3>, CWD <$HOME/.gc3pie_jobs/lrm
                     s_job.YNZmU17755/.>, Output File <lsf.o%J>, Requested Reso
                     urces <select[mem<70000 && lustre] order[-ut] rusage[mem=1
                     000,m=1]>, Login Shell </bin/sh>, Specified Hosts <thin+9>
                     , <single+8>, <smp16+6>, <smp24+5>, <smp48+4>;

 RUNLIMIT
 480.0 min of a6122
Tue Jul 24 10:04:19: Started on <a6122>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
                     obs/lrms_job.YNZmU17755/.>;
Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.

 MEMORY USAGE:
 MAX MEM: 41 Mbytes;  AVG MEM: 41 Mbytes

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -  1000M     -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched 4000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
"""

    # Also parse the output of jobs to get accounting information
    acct = lsf._parse_acct_output(bjobs_output, '')
    assert acct['duration'] == Duration('86s')
    assert acct['used_cpu_time'] == Duration('2.1s')
    assert acct['max_used_memory'] == Memory('41MB')


def test_bjobs_output_done2():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    jobstatus = lsf._parse_stat_output("""
Job <726659>, Job Name <Application>, User <wwolski>, Project <default>, Status
                      <DONE>, Queue <pub.8h>, Job Priority <50>, Command <FileM
                     erger -in /IMSB/users/wwolski/gc3pieScripts/split_napedro_
                     L120224_001_SW-400AQUA_no_background_2ul_dilution.ChromaEx
                     tractShort/split_napedro_L120224_001_SW-400AQUA_no_backgro
                     und_2ul_dilution_10_12._rtnorm.chrom.mzML /IMSB/users/wwol
                     ski/gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_
                     background_2ul_dilution.ChromaExtractShort/split_napedro_L
                     120224_001_SW-400AQUA_no_background_2ul_dilution_10_13._rt
                     norm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split_na
                     pedro_L120224_001_SW-400AQUA_no_background_2ul_dilution.Ch
                     romaExtractShort/split_napedro_L120224_001_SW-400AQUA_no_b
                     ackground_2ul_dilution_10_14._rtnorm.chrom.mzML /IMSB/user
                     s/wwolski/gc3pieScripts/split_napedro_L120224_001_SW-400AQ
                     UA_no_background_2ul_dilution.ChromaExtractShort/split_nap
                     edro_L120224_001_SW-400AQUA_no_background_2ul_dilution_10_
                     15._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/sp
                     lit_napedro_L120224_001_SW-400AQUA_no_background_2ul_dilut
                     ion.ChromaExtractShort/split_napedro_L120224_001_SW-400AQU
                     A_no_background_2ul_dilution_10_16._rtnorm.chrom.mzML /IMS
                     B/users/wwolski/gc3pieScripts/split_napedro_L120224_001_SW
                     -400AQUA_no_background_2ul_dilution.ChromaExtractShort/spl
                     it_napedro_L120224_001_SW-400AQUA_no_background_2ul_diluti
                     on_10_17._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScri
                     pts/split_napedro_L120224_001_SW-400AQUA_no_background_2ul
                     _dilution.ChromaExtractShort/split_napedro_L120224_001_SW-
                     400AQUA_no_background_2ul_dilution_10_18._rtnorm.chrom.mzM
                     L /IMSB/users/wwolski/gc3pieScripts/split_napedro_L120224_
                     001_SW-400AQUA_no_background_2ul_dilution.ChromaExtractSho
                     rt/split_napedro_L120224_001_SW-400AQUA_no_background_2ul_
                     dilution_10_19._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3p
                     ieScripts/split_napedro_L120224_001_SW-400AQUA_no_backgrou
                     nd_2ul_dilution.ChromaExtractShort/split_napedro_L120224_0
                     01_SW-400AQUA_no_background_2ul_dilution_10_1._rtnorm.chro
                     m.mzML /IMSB/users/wwolski/gc3pieScripts/split_napedro_L12
                     0224_001_SW-400AQUA_no_background_2ul_dilution.ChromaExtra
                     ctShort/split_napedro_L120224_001_SW-400AQUA_no_background
                     _2ul_dilution_10_20._rtnorm.chrom.mzML /IMSB/users/wwolski
                     /gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_bac
                     kground_2ul_dilution.ChromaExtractShort/split_napedro_L120
                     224_001_SW-400AQUA_no_background_2ul_dilution_10_21._rtnor
                     m.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split_naped
                     ro_L120224_001_SW-400AQUA_no_background_2ul_dilution.Chrom
                     aExtractShort/split_napedro_L120224_001_SW-400AQUA_no_back
                     ground_2ul_dilution_10_22._rtnorm.chrom.mzML /IMSB/users/w
                     wolski/gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_
                     no_background_2ul_dilution.ChromaExtractShort/split_napedr
                     o_L120224_001_SW-400AQUA_no_background_2ul_dilution_10_23.
                     _rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split
                     _napedro_L120224_001_SW-400AQUA_no_background_2ul_dilution
                     .ChromaExtractShort/split_napedro_L120224_001_SW-400AQUA_n
                     o_background_2ul_dilution_10_24._rtnorm.chrom.mzML /IMSB/u
                     sers/wwolski/gc3pieScripts/split_napedro_L120224_001_SW-40
                     0AQUA_no_background_2ul_dilution.ChromaExtractShort/split_
                     napedro_L120224_001_SW-400AQUA_no_background_2ul_dilution_
                     10_25._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts
                     /split_napedro_L120224_001_SW-400AQUA_no_background_2ul_di
                     lution.ChromaExtractShort/split_napedro_L120224_001_SW-400
                     AQUA_no_background_2ul_dilution_10_26._rtnorm.chrom.mzML /
                     IMSB/users/wwolski/gc3pieScripts/split_napedro_L120224_001
                     _SW-400AQUA_no_background_2ul_dilution.ChromaExtractShort/
                     split_napedro_L120224_001_SW-400AQUA_no_background_2ul_dil
                     ution_10_27._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieS
                     cripts/split_napedro_L120224_001_SW-400AQUA_no_background_
                     2ul_dilution.ChromaExtractShort/split_napedro_L120224_001_
                     SW-400AQUA_no_background_2ul_dilution_10_28._rtnorm.chrom.
                     mzML /IMSB/users/wwolski/gc3pieScripts/split_napedro_L1202
                     24_001_SW-400AQUA_no_background_2ul_dilution.ChromaExtract
                     Short/split_napedro_L120224_001_SW-400AQUA_no_background_2
                     ul_dilution_10_29._rtnorm.chrom.mzML /IMSB/users/wwolski/g
                     c3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_backg
                     round_2ul_dilution.ChromaExtractShort/split_napedro_L12022
                     4_001_SW-400AQUA_no_background>, Share group charged </lsf
                     _biol_all/lsf_biol_other/wwolski>
Mon Jul 30 15:12:05: Submitted from host <brutus2>, CWD <$HOME/.gc3pie_jobs/lrm
                     s_job.QwyFmi4681/.>, Output File <lsf.o%J>, Requested Reso
                     urces <select[mem<70000] order[-ut] rusage[mem=1000,m=1]>,
                      Login Shell </bin/sh>, Specified Hosts <thin+9>, <single+
                     8>, <smp16+6>, <smp24+5>, <smp48+4>;

 RUNLIMIT
 480.0 min of a3168
Mon Jul 30 15:12:47: Started on <a3168>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
                     obs/lrms_job.QwyFmi4681/.>;
Mon Jul 30 15:12:56: Done successfully. The CPU time used is 1.7 seconds.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched20000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
""",
    # STDERR
    '')
    assert jobstatus.state == gc3libs.Run.State.TERMINATING
    assert jobstatus.termstatus == (0, 0)


def test_bjobs_output_done3():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    jobstatus = lsf._parse_stat_output("""
Job <2073>, Job Name <GRunApplication.0>, User <markmon>, Project <default>, St
                          atus <EXIT>, Queue <normal>, Command <sh -c inputfile
                          .txt>
Mon Aug  4 12:28:51 2014: Submitted from host <pa64.dri.edu>, CWD <$HOME/.gc3pi
                          e_jobs/lrms_job.5gDDxlxcty>, Specified CWD <$HOME/.gc
                          3pie_jobs/lrms_job.5gDDxlxcty/.>, Output File (overwr
                          ite) <stdout.txt>, Error File (overwrite) <stderr.txt
                          >, Requested Resources <rusage[mem=2000]>, Login Shel
                          l </bin/sh>;

 RUNLIMIT
 480.0 min of pa54.dri.edu
Mon Aug  4 12:28:51 2014: Started on <pa54.dri.edu>, Execution Home </home/mark
                          mon>, Execution CWD </home/markmon/.gc3pie_jobs/lrms_
                          job.5gDDxlxcty/.>;
Mon Aug  4 12:28:51 2014: Exited with exit code 127. The CPU time used is 0.1 s
                          econds.
Mon Aug  4 12:28:51 2014: Completed <exit>.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

 RESOURCE REQUIREMENT DETAILS:
 Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
 Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
    """,
    # STDERR
    '')
    assert jobstatus.state == gc3libs.Run.State.TERMINATING
    assert jobstatus.termstatus == (0, 127)


def test_bjobs_output_done_long_ago():
    """Test parsing `bjobs -l` output for a job that was removed from `mbatchd` core memory"""
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    jobstatus = lsf._parse_stat_output(
        # empty STDOUT
        '',
        # STDERR
        'Job <943186> is not found')
    assert jobstatus.state == gc3libs.Run.State.TERMINATING
    assert jobstatus.termstatus == None


def test_bjobs_output_exit_nonzero():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local')
    jobstatus = lsf._parse_stat_output("""
Job <132286>, User <wwolski>, Project <default>, Status <EXIT>, Queue <pub.1h>,
                     Job Priority <50>, Command <./x.sh>, Share group charged <
                     /lsf_biol_all/lsf_biol_other/wwolski>
Tue Jul 24 10:26:42: Submitted from host <brutus3>, CWD <$HOME/.>, Output File
                     <lsf.o%J>, Requested Resources <select[mem<70000] order[-u
                     t] rusage[mem=1024,xs=1]>, Specified Hosts <thin+9>, <sing
                     le+8>, <smp16+6>, <smp24+5>, <smp48+4>, <parallel+1>;
Tue Jul 24 10:26:47: Started on <a3010>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.>;
Tue Jul 24 10:26:53: Exited with exit code 42. The CPU time used is 0.0 seconds
                     .
Tue Jul 24 10:26:53: Completed <exit>.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched20000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
    """,
    # STDERR
    '')
    assert jobstatus.state == gc3libs.Run.State.TERMINATING
    assert jobstatus.termstatus == (0, 42)


def test_bjobs_incorrect_prefix_length():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local',
                  lsf_continuation_line_prefix_length=7)
    stat_result = lsf._parse_stat_output("""
Job <2073>, Job Name <GRunApplication.0>, User <markmon>, Project <default>, St
                          atus <EXIT>, Queue <normal>, Command <sh -c inputfile
                          .txt>
Mon Aug  4 12:28:51 2014: Submitted from host <pa64.dri.edu>, CWD <$HOME/.gc3pi
                          e_jobs/lrms_job.5gDDxlxcty>, Specified CWD <$HOME/.gc
                          3pie_jobs/lrms_job.5gDDxlxcty/.>, Output File (overwr
                          ite) <stdout.txt>, Error File (overwrite) <stderr.txt
                          >, Requested Resources <rusage[mem=2000]>, Login Shel
                          l </bin/sh>;

 RUNLIMIT
 480.0 min of pa54.dri.edu
Mon Aug  4 12:28:51 2014: Started on <pa54.dri.edu>, Execution Home </home/mark
                          mon>, Execution CWD </home/markmon/.gc3pie_jobs/lrms_
                          job.5gDDxlxcty/.>;
Mon Aug  4 12:28:51 2014: Exited with exit code 127. The CPU time used is 0.1 s
                          econds.
Mon Aug  4 12:28:51 2014: Completed <exit>.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

 RESOURCE REQUIREMENT DETAILS:
 Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
 Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
""",
    # STDERR
    '')
    assert stat_result.state == gc3libs.Run.State.UNKNOWN
    assert stat_result.termstatus == None


def test_bjobs_correct_explicit_prefix_length():
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local',
                  lsf_continuation_line_prefix_length=26)
    stat_result = lsf._parse_stat_output("""
Job <2073>, Job Name <GRunApplication.0>, User <markmon>, Project <default>, St
                          atus <EXIT>, Queue <normal>, Command <sh -c inputfile
                          .txt>
Mon Aug  4 12:28:51 2014: Submitted from host <pa64.dri.edu>, CWD <$HOME/.gc3pi
                          e_jobs/lrms_job.5gDDxlxcty>, Specified CWD <$HOME/.gc
                          3pie_jobs/lrms_job.5gDDxlxcty/.>, Output File (overwr
                          ite) <stdout.txt>, Error File (overwrite) <stderr.txt
                          >, Requested Resources <rusage[mem=2000]>, Login Shel
                          l </bin/sh>;

 RUNLIMIT
 480.0 min of pa54.dri.edu
Mon Aug  4 12:28:51 2014: Started on <pa54.dri.edu>, Execution Home </home/mark
                          mon>, Execution CWD </home/markmon/.gc3pie_jobs/lrms_
                          job.5gDDxlxcty/.>;
Mon Aug  4 12:28:51 2014: Exited with exit code 127. The CPU time used is 0.1 s
                          econds.
Mon Aug  4 12:28:51 2014: Completed <exit>.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

 RESOURCE REQUIREMENT DETAILS:
 Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
 Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
""",
    # STDERR
    '')
    assert stat_result.state == gc3libs.Run.State.TERMINATING
    assert stat_result.termstatus == (0, 127)


def test_bacct_done0():
    """Test parsing accounting information of a <sleep 300> job."""
    # gotten with `bacct -l "jobid"`
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local',
                  bacct='bacct')
    acct = lsf._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3329613>, User <rmurri>, Project <default>, Status <DONE>, Queue <pub.1h>,
                     Command <sleep 60>, Share group charged </lsf_biol_all/lsf
                     _aeber/rmurri>
Mon Oct  8 17:07:54: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Mon Oct  8 17:08:44: Dispatched to <a3201>;
Mon Oct  8 17:09:51: Completed <done>.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.08       50            117     done         0.0007     5M    222M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       1      Total number of exited jobs:     0
 Total CPU time consumed:       0.1      Average CPU time consumed:     0.1
 Maximum CPU time of a job:     0.1      Minimum CPU time of a job:     0.1
 Total wait time in queues:    50.0
 Average wait time in queue:   50.0
 Maximum wait time in queue:   50.0      Minimum wait time in queue:   50.0
 Average turnaround time:       117 (seconds/job)
 Maximum turnaround time:       117      Minimum turnaround time:       117
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00

    """,
    # STDERR
    '')
    assert acct['duration'] == Duration('67s')
    assert acct['used_cpu_time'] == Duration('0.08s')
    assert acct['max_used_memory'] == Memory('227MB')
    # timestamps
    year = datetime.date.today().year
    assert (acct['lsf_submission_time'] ==
                 datetime.datetime(year, 10, 8, 17, 7, 54))
    assert (acct['lsf_start_time'] ==
                 datetime.datetime(year, 10, 8, 17, 8, 44))
    assert (acct['lsf_completion_time'] ==
                 datetime.datetime(year, 10, 8, 17, 9, 51))


def test_bacct_done1():
    """Test parsing `bacct -l` output for a not-so-trivial job."""
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local',
                  bacct='bacct')
    acct = lsf._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3329618>, User <rmurri>, Project <default>, Status <DONE>, Queue <pub.1h>,
                     Command <md5sum lsf.o3224113 lsf.o3224132>, Share group ch
                     arged </lsf_biol_all/lsf_aeber/rmurri>
Mon Oct  8 17:08:54: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Mon Oct  8 17:10:01: Dispatched to <a3041>;
Mon Oct  8 17:10:07: Completed <done>.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.04       67             73     done         0.0005     3M     34M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       1      Total number of exited jobs:     0
 Total CPU time consumed:       0.0      Average CPU time consumed:     0.0
 Maximum CPU time of a job:     0.0      Minimum CPU time of a job:     0.0
 Total wait time in queues:    67.0
 Average wait time in queue:   67.0
 Maximum wait time in queue:   67.0      Minimum wait time in queue:   67.0
 Average turnaround time:        73 (seconds/job)
 Maximum turnaround time:        73      Minimum turnaround time:        73
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
    """,
    # STDERR
    '')
    assert acct['duration'] == Duration('6s')
    assert acct['used_cpu_time'] == Duration('0.04s')
    assert acct['max_used_memory'] == Memory('37MB')
    # timestamps
    year = datetime.date.today().year
    assert (acct['lsf_submission_time'] ==
                 datetime.datetime(year, 10, 8, 17, 8, 54))
    assert (acct['lsf_start_time'] ==
                 datetime.datetime(year, 10, 8, 17, 10, 1))
    assert (acct['lsf_completion_time'] ==
                 datetime.datetime(year, 10, 8, 17, 10, 7))


def test_bacct_killed():
    """Test parsing `bacct -l` output for a canceled job."""
    lsf = LsfLrms(name='test',
                  architecture=gc3libs.Run.Arch.X86_64,
                  max_cores=1,
                  max_cores_per_job=1,
                  max_memory_per_core=1 * GB,
                  max_walltime=1 * hours,
                  auth=None,  # ignored if `transport` is `local`
                  frontend='localhost',
                  transport='local',
                  bacct='bacct')
    acct = lsf._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3224113>, User <rmurri>, Project <default>, Status <EXIT>, Queue <pub.1h>,
                     Command <sleep 300>, Share group charged </lsf_biol_all/ls
                     f_aeber/rmurri>
Fri Oct  5 17:49:35: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Fri Oct  5 17:50:35: Dispatched to <a3191>;
Fri Oct  5 17:51:30: Completed <exit>; TERM_OWNER: job killed by owner.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.04       60            115     exit         0.0003     1M     34M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       0      Total number of exited jobs:     1
 Total CPU time consumed:       0.0      Average CPU time consumed:     0.0
 Maximum CPU time of a job:     0.0      Minimum CPU time of a job:     0.0
 Total wait time in queues:    60.0
 Average wait time in queue:   60.0
 Maximum wait time in queue:   60.0      Minimum wait time in queue:   60.0
 Average turnaround time:       115 (seconds/job)
 Maximum turnaround time:       115      Minimum turnaround time:       115
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
""",
    # STDERR
    '')
    assert acct['duration'] == Duration('55s')
    assert acct['used_cpu_time'] == Duration('0.04s')
    assert acct['max_used_memory'] == Memory('35MB')
    # timestamps
    year = datetime.date.today().year
    assert (acct['lsf_submission_time'] ==
                 datetime.datetime(year, 10, 5, 17, 49, 35))
    assert (acct['lsf_start_time'] ==
                 datetime.datetime(year, 10, 5, 17, 50, 35))
    assert (acct['lsf_completion_time'] ==
                 datetime.datetime(year, 10, 5, 17, 51, 30))


# LSF incorporates resource usage information in a job's output;
# the job's output is a copy of the email that the LSF system
# sends to the user that submitted a job.
###############################################################################
# Sender: LSF System <lsfadmin@cpt086>
# Subject: Job 943186: <md5sum XXX/x.sh> Done
#
# Job <md5sum XXX/x.sh> was submitted from host <frt> by user <rmurri>
# in cluster <prdclst>.  Job was executed on host(s) <cpt086>, in
# queue <normal>, as user <rmurri> in cluster <prdclst>.
# </home/rmurri> was used as the home directory.  </home/rmurri> was
# used as the working directory.  Started at Mon Oct 8 17:23:30 2012
# Results reported at Mon Oct 8 17:23:31 2012
#
# Your job looked like:
#
# ------------------------------------------------------------
# LSBATCH: User input
# md5sum XXX/x.sh
# ------------------------------------------------------------
#
# Successfully completed.
#
# Resource usage summary:
#
#     CPU time   :      0.01 sec.
#     Max Memory :         2 MB
#     Max Swap   :        24 MB
#
#     Max Processes  :         1
#     Max Threads    :         1
#
# The output (if any) follows:
#
# b7f9d9c86469aa8b57c19af14bf80af9  XXX/x.sh
###############################################################################


if __name__ == "__main__":
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gprecovery.py -- Front-end script for running ParRecoveryFun Matlab
#   function with a given combination of reference models.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gprecovery.py --help`` for program usage
instructions.

Input parameters consists of:

...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-03-14:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gprecovery
    gprecovery.GprecoveryScript().run()

import os
import sys
import time
import tempfile

import shutil
import random

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

MODELS_SPECS = [1,2,3,4]
## custom application class
class GprecoveryApplication(Application):
    """
    Custom class to wrap the execution of the Matlab script.
    """
    application_name = 'matlab-mcr'

    def __init__(self, model_index, seed, **extra_args):

        self.output_dir = extra_args['output_dir']

        self.result_dir = extra_args['result_dir']

        inputs = dict()
        outputs = dict()
        executables = []

        if 'run_binary' in extra_args:
            inputs[os.path.abspath(extra_args['run_binary'])] = './par_recovery'
            arguments = "./par_recovery "
            executables.append('./par_recovery')
        else:
            arguments = "par_recovery "

        # self.output_filename = "ParRecovery_Genmodel%s_.mat" % str(model_index)
        # outputs[self.output_filename] = os.path.join(self.result_dir, "ParRecovery_Genmodel%s_%s.mat" % (str(model_index),extra_args['repetition']))
        self.output_filename = "ParRecovery_Genmodel%s_%s.mat" % (str(model_index),extra_args['repetition'])
        outputs["ParRecovery_Genmodel%s_.mat" % str(model_index)] = self.output_filename

        arguments += "%s %s" % (str(model_index), str(seed))

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gprecovery.log',
            join=True,
            executables = executables,
            **extra_args)


    def terminated(self):
        """
        Move output file in 'result_dir'
        """
        if os.path.isfile(os.path.join(self.output_dir,self.output_filename)):
            shutil.move(os.path.join(self.output_dir,self.output_filename),
                             os.path.join(self.result_dir,self.output_filename))
        else:
            gc3libs.log.error("Expected output file %s not found."
                              % os.path.join(self.output_dir,self.output_filename))


class GprecoveryScript(SessionBasedScript):
    """
    Fro each network file (with '.dat' extension) found in the 'input folder',
    GprecoveryScript generates as many Tasks as 'benchmarks' defined.

    The ``gprecovery`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gprecovery`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GprecoveryApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GprecoveryApplication,
            )

    def setup_args(self):

        self.add_param('models', type=str,
                       help="list of models to process. Valid ranges are: %s."
                       " Syntax allowed: "
                       " [index|index_start:index_end|index_1,...,index_N]."
                       " Es. 1,3 | 1:4 | 3" % MODELS_SPECS)

    def setup_options(self):
        self.add_param("-b", "--binary", metavar="[STRING]",
                       dest="run_binary", default=None,
                       help="Location of the Matlab compiled binary "
                       "version of the ParRecoveryFun. Default: None.")

        self.add_param("-E", "--random_range", type=int, metavar="[int]",
                       dest="random_range", default=1000,
                       help="Upper limit for the random seed used in the "
                       "fmin function. Default: 1000.")

        self.add_param("-R", "--repeat", type=int, metavar="[int]",
                       dest="repeat", default=1,
                       help="Repeat all simulation [repeat] times. "
                       " Default: 1 (no repeat).")

        self.add_param("-S", "--store_results", type=str, metavar="[STRING]",
                       dest="store_results", default=None,
                       help="Location where all results will be aggregated. "
                       "Default: (session folder).")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        self.models = []

        if self.params.run_binary:
            if not os.path.isfile(self.params.run_binary):
                raise gc3libs.exceptions.InvalidUsage("ParRecoveryFun binary "
                                                      " file %s not found"
                                                      % self.params.run_binary)
        try:
            if self.params.models.count(':') == 1:
                start, end = self.params.models.split(':')
                if (int(start) <= int(end)) and (int(start) in MODELS_SPECS) and (int(end) in MODELS_SPECS):
                    self.models = range(int(start), int(end)+1)
                else:
                    raise gc3libs.exceptions.InvalidUsage(
                        "Model not in valid range. "
                        "Range: %s" % str(MODELS_SPECS))
            elif self.params.models.count(',') >= 1:
                self.models = [ int(s) for s in self.params.models.split(',')
                                if int(s) in MODELS_SPECS ]
            else:
                if int(self.params.models) in MODELS_SPECS:
                    self.models = [ int(self.params.models) ]
                else:
                    gc3libs.log.error("Model %s not in valid range. "
                                      "Range: %s" % (self.params.models,
                                                     str(MODELS_SPECS)))

        except ValueError:
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid argument '%s', use on of the following formats: "
                " INT:INT | INT,INT,INT,..,INT | INT " % (models,))


    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GprecoveryApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for model in self.models:

            for repeat in range(1,(self.params.repeat + 1)):

                # XXX: need to find a more compact name
                # jobname = "gprecovery-%s-%s" % (model,data_index)
                jobname = "gprecovery-%d-%d" % (model,repeat)

                seed = random.randint(1,self.params.random_range)

                extra_args = extra.copy()

                if self.params.run_binary:
                    extra_args['run_binary'] = self.params.run_binary

                extra_args['jobname'] = jobname
                extra_args['repetition'] = repeat

                if self.params.store_results:
                    if not os.path.isdir(self.params.store_results):
                        os.makedirs(self.params.store_results)
                    extra_args['result_dir'] = self.params.store_results
                else:
                    extra_args['result_dir'] = self.params.output
                    extra_args['result_dir'] = extra_args['result_dir'].replace('NAME', self.params.session)

                extra_args['output_dir'] = self.params.output
                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', '%s.%s' % (str(model),str(repeat)))
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', '%s.%s' % (str(model),str(repeat)))
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', '%s.%s' % (str(model),str(repeat)))
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', '%s.%s' % (str(model),str(repeat)))


                tasks.append(GprecoveryApplication(
                    model,
                    seed,
                    **extra_args))

        return tasks
